CONVERSATION = """
**Role**: You are a help agent that called VisionAent, built by LandingAI, that assists users write code to solve vision tasks.

**Actions**:
`generate_or_edit_vision_code` - This action will generate code for the user to solve a vision task. It will also edit vision code for the user, this is useful if the user wants to modify vision-related aspects of the code such as changing the vision model or the image pre-processing steps.

`edit_code` - If the user already has code for a vision task, from a previous call to `generate_or_edit_vision_code` and wants to quickly modify the code, you can use this action to edit the code. This is good for non-vision related changes such as formatting the response as a json or removing print statements.

**Exampels**:
Here is an example of how you can interact with a user and Actions:
--- EXAMPLE1 ---
USER: Can you help me write code to detect dogs in this image? Media name images/dogs.jpg

AGENT: <response>Yes, I can help you with that. I will write the code to detect dogs in the image.</response><action>generate_or_edit_vision_code</action>

OBSERVATION:
<final_code>
from vision_agent.tools import load_image, owl_v2_image
def detect_dogs(image_path: str):
    image = load_image(image_path)
    dogs = owl_v2_image(image)
    return dogs
</final_code>
<final_test>
def test_detect_dogs():
    dogs = detect_dogs("images/dogs.jpg")
    assert len(dogs) > 0
</final_test>

AGENT: <response>Here is the code to detect dogs in the image.</response>
--- END EXAMPLE1 ---

--- EXAMPLE2 ---
USER: Who are you and what can you do?

AGENT: <response>I am VisionAgent, an agent built by LandingAI, to help users write code to solve vision tasks. I can write code to do the following tasks:

- Detecting objects in an image with bounding boxes
- Detecting objects in an image with masks
- Tracking object in a video
- Classifying images
- Understanding documents
- Pose estimation
- Visual question answering for both images and videos
- Action recognition in videos
- Image inpainting

How can I help you?</response>
--- END EXAMPLE2 ---

**Conversation**:
Here is the current conversation so far:
--- START CONVERSATION ---
{conversation}
--- END CONVERSATION ---

**Instructions**:
1. Only respond with a single <response> tag and a single <action> tag.
2. You can only take one action at a time in response to the user's message. Do not offer to fix code on the user's behalf, only if they have directly asked you to.
3. Respond in the following format, the <action> tag is optional and can be excluded if you do not want to take any action:

<response>Your response to the user's message</response>
<action>The action you want to take from **Actions**</action>
"""



import json
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Sequence, Union, cast

import anthropic
import requests
from anthropic.types import ImageBlockParam, MessageParam, TextBlockParam
from openai import AzureOpenAI, OpenAI

from vision_agent.utils.image_utils import encode_media

from .types import Message


class LMM(ABC):
    @abstractmethod
    def generate(
        self,
        prompt: str,
        media: Optional[Sequence[Union[str, Path]]] = None,
        **kwargs: Any,
    ) -> Union[str, Iterator[Optional[str]]]:
        pass

    @abstractmethod
    def chat(
        self,
        chat: Sequence[Message],
        **kwargs: Any,
    ) -> Union[str, Iterator[Optional[str]]]:
        pass

    @abstractmethod
    def __call__(
        self,
        input: Union[str, Sequence[Message]],
        **kwargs: Any,
    ) -> Union[str, Iterator[Optional[str]]]:
        pass


class OpenAILMM(LMM):
    r"""An LMM class for the OpenAI LMMs."""

    def __init__(
        self,
        model_name: str = "gpt-4o-2024-05-13",
        api_key: Optional[str] = None,
        max_tokens: int = 4096,
        json_mode: bool = False,
        image_size: int = 768,
        image_detail: str = "low",
        **kwargs: Any,
    ):
        if not api_key:
            self.client = OpenAI()
        else:
            self.client = OpenAI(api_key=api_key)

        self.client = OpenAI(api_key=api_key)
        self.model_name = model_name
        self.image_size = image_size
        self.image_detail = image_detail
        # o1 does not use max_tokens
        if "max_tokens" not in kwargs and not model_name.startswith("o1"):
            kwargs["max_tokens"] = max_tokens
        if json_mode:
            kwargs["response_format"] = {"type": "json_object"}
        self.kwargs = kwargs

    def __call__(
        self,
        input: Union[str, Sequence[Message]],
        **kwargs: Any,
    ) -> Union[str, Iterator[Optional[str]]]:
        if isinstance(input, str):
            return self.generate(input, **kwargs)
        return self.chat(input, **kwargs)

    def chat(
        self,
        chat: Sequence[Message],
        **kwargs: Any,
    ) -> Union[str, Iterator[Optional[str]]]:
        """Chat with the LMM model.

        Parameters:
            chat (Squence[Dict[str, str]]): A list of dictionaries containing the chat
                messages. The messages can be in the format:
                [{"role": "user", "content": "Hello!"}, ...]
                or if it contains media, it should be in the format:
                [{"role": "user", "content": "Hello!", "media": ["image1.jpg", ...]}, ...]
        """
        fixed_chat = []
        for c in chat:
            fixed_c = {"role": c["role"]}
            fixed_c["content"] = [{"type": "text", "text": c["content"]}]  # type: ignore
            if "media" in c:
                for media in c["media"]:
                    resize = kwargs["resize"] if "resize" in kwargs else self.image_size
                    image_detail = (
                        kwargs["image_detail"]
                        if "image_detail" in kwargs
                        else self.image_detail
                    )
                    encoded_media = encode_media(cast(str, media), resize=resize)

                    fixed_c["content"].append(  # type: ignore
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": (
                                    encoded_media
                                    if encoded_media.startswith(("http", "https"))
                                    or encoded_media.startswith("data:image/")
                                    else f"data:image/png;base64,{encoded_media}"
                                ),
                                "detail": image_detail,
                            },
                        },
                    )
            fixed_chat.append(fixed_c)

        # prefers kwargs from second dictionary over first
        tmp_kwargs = self.kwargs | kwargs
        response = self.client.chat.completions.create(
            model=self.model_name, messages=fixed_chat, **tmp_kwargs  # type: ignore
        )
        if "stream" in tmp_kwargs and tmp_kwargs["stream"]:

            def f() -> Iterator[Optional[str]]:
                for chunk in response:
                    chunk_message = chunk.choices[0].delta.content  # type: ignore
                    yield chunk_message

            return f()
        else:
            return cast(str, response.choices[0].message.content)

    def generate(
        self,
        prompt: str,
        media: Optional[Sequence[Union[str, Path]]] = None,
        **kwargs: Any,
    ) -> Union[str, Iterator[Optional[str]]]:
        message: List[Dict[str, Any]] = [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": prompt},
                ],
            }
        ]
        if media and len(media) > 0:
            for m in media:
                resize = kwargs["resize"] if "resize" in kwargs else None
                image_detail = (
                    kwargs["image_detail"]
                    if "image_detail" in kwargs
                    else self.image_detail
                )
                encoded_media = encode_media(m, resize=resize)
                message[0]["content"].append(
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": (
                                encoded_media
                                if encoded_media.startswith(("http", "https"))
                                or encoded_media.startswith("data:image/")
                                else f"data:image/png;base64,{encoded_media}"
                            ),
                            "detail": image_detail,
                        },
                    },
                )

        # prefers kwargs from second dictionary over first
        tmp_kwargs = self.kwargs | kwargs
        response = self.client.chat.completions.create(
            model=self.model_name, messages=message, **tmp_kwargs  # type: ignore
        )
        if "stream" in tmp_kwargs and tmp_kwargs["stream"]:

            def f() -> Iterator[Optional[str]]:
                for chunk in response:
                    chunk_message = chunk.choices[0].delta.content  # type: ignore
                    yield chunk_message

            return f()
        else:
            return cast(str, response.choices[0].message.content)


class AzureOpenAILMM(OpenAILMM):
    def __init__(
        self,
        model_name: Optional[str] = None,
        api_key: Optional[str] = None,
        api_version: str = "2024-02-01",
        azure_endpoint: Optional[str] = None,
        max_tokens: int = 4096,
        json_mode: bool = False,
        image_detail: str = "low",
        **kwargs: Any,
    ):
        if not api_key:
            api_key = os.getenv("AZURE_OPENAI_API_KEY")
        if not azure_endpoint:
            azure_endpoint = os.getenv("AZURE_OPENAI_ENDPOINT")
        if not model_name:
            model_name = os.getenv("AZURE_OPENAI_CHAT_MODEL_DEPLOYMENT_NAME")

        if not api_key:
            raise ValueError("OpenAI API key is required.")
        if not azure_endpoint:
            raise ValueError("Azure OpenAI endpoint is required.")
        if not model_name:
            raise ValueError("Azure OpenAI chat model deployment name is required.")

        self.client = AzureOpenAI(
            api_key=api_key,
            api_version=api_version,
            azure_endpoint=azure_endpoint,
        )
        self.model_name = model_name
        self.image_detail = image_detail

        if "max_tokens" not in kwargs:
            kwargs["max_tokens"] = max_tokens
        if json_mode:
            kwargs["response_format"] = {"type": "json_object"}
        self.kwargs = kwargs


class OllamaLMM(LMM):
    r"""An LMM class for the ollama."""

    def __init__(
        self,
        model_name: str = "llava",
        base_url: Optional[str] = "http://localhost:11434/api",
        json_mode: bool = False,
        num_ctx: int = 128_000,
        image_size: int = 768,
        **kwargs: Any,
    ):
        """Initializes the Ollama LMM. kwargs are passed as 'options' to the model.
        More information on options can be found here
        https://github.com/ollama/ollama/blob/main/docs/modelfile.md#valid-parameters-and-values

        Parameters:
            model_name (str): The ollama name of the model.
            base_url (str): The base URL of the Ollama API.
            json_mode (bool): Whether to use JSON mode.
            num_ctx (int): The context length for the model.
            kwargs (Any): Additional options to pass to the model.
        """

        self.url = base_url
        self.model_name = model_name
        self.image_size = image_size
        self.kwargs = {"options": kwargs}

        if json_mode:
            self.kwargs["format"] = "json"  # type: ignore
        self.kwargs["options"]["num_cxt"] = num_ctx

    def __call__(
        self,
        input: Union[str, Sequence[Message]],
        **kwargs: Any,
    ) -> Union[str, Iterator[Optional[str]]]:
        if isinstance(input, str):
            return self.generate(input, **kwargs)
        return self.chat(input, **kwargs)

    def chat(
        self,
        chat: Sequence[Message],
        **kwargs: Any,
    ) -> Union[str, Iterator[Optional[str]]]:
        """Chat with the LMM model.

        Parameters:
            chat (Sequence[Dict[str, str]]): A list of dictionaries containing the chat
                messages. The messages can be in the format:
                [{"role": "user", "content": "Hello!"}, ...]
                or if it contains media, it should be in the format:
                [{"role": "user", "content": "Hello!", "media": ["image1.jpg", ...]}, ...]
        """
        fixed_chat = []
        for message in chat:
            if "media" in message:
                resize = kwargs["resize"] if "resize" in kwargs else self.image_size
                message["images"] = [
                    encode_media(cast(str, m), resize=resize) for m in message["media"]
                ]
                del message["media"]
            fixed_chat.append(message)
        url = f"{self.url}/chat"
        model = self.model_name
        messages = fixed_chat
        data: Dict[str, Any] = {"model": model, "messages": messages}

        tmp_kwargs = self.kwargs | kwargs
        data.update(tmp_kwargs)
        if "stream" in tmp_kwargs and tmp_kwargs["stream"]:
            json_data = json.dumps(data)

            def f() -> Iterator[Optional[str]]:
                with requests.post(url, data=json_data, stream=True) as stream:
                    if stream.status_code != 200:
                        raise ValueError(
                            f"Request failed with status code {stream.status_code}"
                        )

                    for chunk in stream.iter_content(chunk_size=None):
                        chunk_data = json.loads(chunk)
                        if chunk_data["done"]:
                            yield None
                        else:
                            yield chunk_data["message"]["content"]

            return f()
        else:
            data["stream"] = False
            json_data = json.dumps(data)
            resp = requests.post(url, data=json_data)

            if resp.status_code != 200:
                raise ValueError(f"Request failed with status code {resp.status_code}")
            resp = resp.json()
            return resp["message"]["content"]  # type: ignore

    def generate(
        self,
        prompt: str,
        media: Optional[Sequence[Union[str, Path]]] = None,
        **kwargs: Any,
    ) -> Union[str, Iterator[Optional[str]]]:
        url = f"{self.url}/generate"
        data: Dict[str, Any] = {
            "model": self.model_name,
            "prompt": prompt,
            "images": [],
        }

        if media and len(media) > 0:
            for m in media:
                resize = kwargs["resize"] if "resize" in kwargs else self.image_size
                data["images"].append(encode_media(m, resize=resize))

        tmp_kwargs = self.kwargs | kwargs
        data.update(tmp_kwargs)
        if "stream" in tmp_kwargs and tmp_kwargs["stream"]:
            json_data = json.dumps(data)

            def f() -> Iterator[Optional[str]]:
                with requests.post(url, data=json_data, stream=True) as stream:
                    if stream.status_code != 200:
                        raise ValueError(
                            f"Request failed with status code {stream.status_code}"
                        )

                    for chunk in stream.iter_content(chunk_size=None):
                        chunk_data = json.loads(chunk)
                        if chunk_data["done"]:
                            yield None
                        else:
                            yield chunk_data["response"]

            return f()
        else:
            data["stream"] = False
            json_data = json.dumps(data)
            resp = requests.post(url, data=json_data)

            if resp.status_code != 200:
                raise ValueError(f"Request failed with status code {resp.status_code}")

            resp = resp.json()
            return resp["response"]  # type: ignore


class AnthropicLMM(LMM):
    r"""An LMM class for Anthropic's LMMs."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        model_name: str = "claude-3-5-sonnet-20240620",
        max_tokens: int = 4096,
        image_size: int = 768,
        **kwargs: Any,
    ):
        self.client = anthropic.Anthropic(api_key=api_key)
        self.image_size = image_size
        self.model_name = model_name
        if "max_tokens" not in kwargs:
            kwargs["max_tokens"] = max_tokens
        self.kwargs = kwargs

    def __call__(
        self,
        input: Union[str, Sequence[Dict[str, Any]]],
        **kwargs: Any,
    ) -> Union[str, Iterator[Optional[str]]]:
        if isinstance(input, str):
            return self.generate(input, **kwargs)
        return self.chat(input, **kwargs)

    def chat(
        self,
        chat: Sequence[Dict[str, Any]],
        **kwargs: Any,
    ) -> Union[str, Iterator[Optional[str]]]:
        messages: List[MessageParam] = []
        for msg in chat:
            content: List[Union[TextBlockParam, ImageBlockParam]] = [
                TextBlockParam(type="text", text=msg["content"])
            ]
            if "media" in msg:
                for media_path in msg["media"]:
                    resize = kwargs["resize"] if "resize" in kwargs else self.image_size
                    encoded_media = encode_media(media_path, resize=resize)
                    if encoded_media.startswith("data:image/png;base64,"):
                        encoded_media = encoded_media[len("data:image/png;base64,") :]
                    content.append(
                        ImageBlockParam(
                            type="image",
                            source={
                                "type": "base64",
                                "media_type": "image/png",
                                "data": encoded_media,
                            },
                        )
                    )
            messages.append({"role": msg["role"], "content": content})

        # prefers kwargs from second dictionary over first
        tmp_kwargs = self.kwargs | kwargs
        response = self.client.messages.create(
            model=self.model_name, messages=messages, **tmp_kwargs
        )
        if "stream" in tmp_kwargs and tmp_kwargs["stream"]:

            def f() -> Iterator[Optional[str]]:
                for chunk in response:
                    if (
                        chunk.type == "message_start"
                        or chunk.type == "content_block_start"
                    ):
                        continue
                    elif chunk.type == "content_block_delta":
                        yield chunk.delta.text
                    elif chunk.type == "message_stop":
                        yield None

            return f()
        else:
            return cast(str, response.content[0].text)

    def generate(
        self,
        prompt: str,
        media: Optional[Sequence[Union[str, Path]]] = None,
        **kwargs: Any,
    ) -> Union[str, Iterator[Optional[str]]]:
        content: List[Union[TextBlockParam, ImageBlockParam]] = [
            TextBlockParam(type="text", text=prompt)
        ]
        if media:
            for m in media:
                resize = kwargs["resize"] if "resize" in kwargs else self.image_size
                encoded_media = encode_media(m, resize=resize)
                if encoded_media.startswith("data:image/png;base64,"):
                    encoded_media = encoded_media[len("data:image/png;base64,") :]
                content.append(
                    ImageBlockParam(
                        type="image",
                        source={
                            "type": "base64",
                            "media_type": "image/png",
                            "data": encoded_media,
                        },
                    )
                )

        # prefers kwargs from second dictionary over first
        tmp_kwargs = self.kwargs | kwargs
        response = self.client.messages.create(
            model=self.model_name,
            messages=[{"role": "user", "content": content}],
            **tmp_kwargs,
        )
        if "stream" in tmp_kwargs and tmp_kwargs["stream"]:

            def f() -> Iterator[Optional[str]]:
                for chunk in response:
                    if (
                        chunk.type == "message_start"
                        or chunk.type == "content_block_start"
                    ):
                        continue
                    elif chunk.type == "content_block_delta":
                        yield chunk.delta.text
                    elif chunk.type == "message_stop":
                        yield None

            return f()
        else:
            return cast(str, response.content[0].text)


class GoogleLMM(OpenAILMM):
    r"""An LMM class for the Google LMMs."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        model_name: str = "gemini-2.0-flash-exp",
        max_tokens: int = 4096,
        image_detail: str = "low",
        image_size: int = 768,
        **kwargs: Any,
    ):
        base_url = "https://generativelanguage.googleapis.com/v1beta/openai/"
        if not api_key:
            api_key = os.environ.get("GEMINI_API_KEY")

        self.client = OpenAI(api_key=api_key, base_url=base_url)

        self.model_name = model_name
        self.image_size = image_size
        self.image_detail = image_detail

        if "max_tokens" not in kwargs:
            kwargs["max_tokens"] = max_tokens
        self.kwargs = kwargs



USER_REQ = """
## User Request
{user_request}
"""

PLAN = """
**Context**:
{context}

**Tools Available**:
{tool_desc}

**Previous Feedback**:
{feedback}

**Instructions**:
1. Based on the context and tools you have available, create a plan of subtasks to achieve the user request.
2. For each subtask, be sure to include the tool(s) you want to use to accomplish that subtask.
3. Output three different plans each utilize a different strategy or set of tools ordering them from most likely to least likely to succeed.

Output a list of jsons in the following format:

```json
{{
    "plan1":
        {{
            "thoughts": str # your thought process for choosing this plan
            "instructions": [
                str # what you should do in this task associated with a tool
            ]
        }},
    "plan2": ...,
    "plan3": ...
}}
```
"""

TEST_PLANS = """
**Role**: You are a software programmer responsible for testing different tools.

**Task**: Your responsibility is to take a set of several plans and test the different tools for each plan.

**Documentation**:
This is the documentation for the functions you have access to. You may call any of these functions to help you complete the task. They are available through importing `from vision_agent.tools import *`.

{docstring}

**Plans**:
{plans}

**Previous Attempts**:
{previous_attempts}

**Examples**:
--- EXAMPLE1 ---
plan1:
- Load the image from the provided file path 'image.jpg'.
- Use the 'owlv2_object_detection' tool with the prompt 'person' to detect and count the number of people in the image.
plan2:
- Load the image from the provided file path 'image.jpg'.
- Use the 'florence2_sam2_instance_segmentation' tool with the prompt 'person' to detect and count the number of people in the image.
- Count the number of detected objects labeled as 'person'.
plan3:
- Load the image from the provided file path 'image.jpg'.
- Use the 'countgd_object_detection' tool to count the dominant foreground object, which in this case is people.

```python
from vision_agent.tools import load_image, owlv2_object_detection, florence2_sam2_instance_segmentation, countgd_object_detection
image = load_image("image.jpg")
owl_v2_out = owlv2_object_detection("person", image)

f2s2_out = florence2_sam2_instance_segmentation("person", image)
# strip out the masks from the output becuase they don't provide useful information when printed
f2s2_out = [{{k: v for k, v in o.items() if k != "mask"}} for o in f2s2_out]

cgd_out = countgd_object_detection("person", image)

final_out = {{"owlv2_object_detection": owl_v2_out, "florence2_sam2_instance_segmentation": f2s2, "countgd_object_detection": cgd_out}}
print(final_out)
--- END EXAMPLE1 ---

--- EXAMPLE2 ---
plan1:
- Extract frames from 'video.mp4' at 10 FPS using the 'extract_frames_and_timestamps' tool.
- Use the 'owl_v2_video' tool with the prompt 'person' to detect where the people are in the video.
plan2:
- Extract frames from 'video.mp4' at 10 FPS using the 'extract_frames_and_timestamps' tool.
- Use the 'florence2_phrase_grounding' tool with the prompt 'person' to detect where the people are in the video.
plan3:
- Extract frames from 'video.mp4' at 10 FPS using the 'extract_frames_and_timestamps' tool.
- Use the 'florence2_sam2_video_tracking' tool with the prompt 'person' to detect where the people are in the video.


```python
import numpy as np
from vision_agent.tools import extract_frames_and_timestamps, owl_v2_video, florence2_phrase_grounding, florence2_sam2_video_tracking

# sample at 1 FPS and use the first 10 frames to reduce processing time
frames = extract_frames_and_timestamps("video.mp4", 1)
frames = [f["frame"] for f in frames][:10]

# strip arrays from the output to make it easier to read
def remove_arrays(o):
    if isinstance(o, list):
        return [remove_arrays(e) for e in o]
    elif isinstance(o, dict):
        return {{k: remove_arrays(v) for k, v in o.items()}}
    elif isinstance(o, np.ndarray):
        return "array: " + str(o.shape)
    else:
        return o

# return the counts of each label per frame to help determine the stability of the model results
def get_counts(preds):
    counts = {{}}
    for i, pred_frame in enumerate(preds):
        counts_i = {{}}
        for pred in pred_frame:
            label = pred["label"].split(":")[1] if ":" in pred["label"] else pred["label"]
            counts_i[label] = counts_i.get(label, 0) + 1
        counts[f"frame_{{i}}"] = counts_i
    return counts


# plan1
owl_v2_out = owl_v2_video("person", frames)
owl_v2_counts = get_counts(owl_v2_out)

# plan2
florence2_out = [florence2_phrase_grounding("person", f) for f in frames]
florence2_counts = get_counts(florence2_out)

# plan3
f2s2_tracking_out = florence2_sam2_video_tracking("person", frames)
remove_arrays(f2s2_tracking_out)
f2s2_counts = get_counts(f2s2_tracking_out)

final_out = {{
    "owl_v2_video": owl_v2_out,
    "florence2_phrase_grounding": florence2_out,
    "florence2_sam2_video_tracking": f2s2_out,
}}

counts = {{
    "owl_v2_video": owl_v2_counts,
    "florence2_phrase_grounding": florence2_counts,
    "florence2_sam2_video_tracking": f2s2_counts,
}}

print(final_out)
print(labels_and_scores)
print(counts)
```
--- END EXAMPLE2 ---

**Instructions**:
1. Write a program to load the media and call each tool and print it's output along with other relevant information.
2. Create a dictionary where the keys are the tool name and the values are the tool outputs. Remove numpy arrays from the printed dictionary.
3. Your test case MUST run only on the given images which are {media}
4. Print this final dictionary.
5. For video input, sample at 1 FPS and use the first 10 frames only to reduce processing time.
"""

PREVIOUS_FAILED = """
**Previous Failed Attempts**:
You previously ran this code:
```python
{code}
```

But got the following error or no stdout:
{error}
"""

PICK_PLAN = """
**Role**: You are an advanced AI model that can understand the user request and construct plans to accomplish it.

**Task**: Your responsibility is to pick the best plan from the three plans provided.

**Context**:
{context}

**Plans**:
{plans}

**Tool Output**:
{tool_output}

**Instructions**:
1. Re-read the user request, plans, tool outputs and examine the image.
2. Solve the problem yourself given the image and pick the most accurate plan that matches your solution the best.
3. Add modifications to improve the plan including: changing a tool, adding thresholds, string matching.
4. Output a JSON object with the following format:
{{
    "predicted_answer": str # the answer you would expect from the best plan
    "thoughts": str # your thought process for choosing the best plan over other plans and any modifications you made
    "best_plan": str # the best plan you have chosen, must be `plan1`, `plan2`, or `plan3`
}}
"""



import inspect
import logging
import os
from base64 import b64encode
from typing import Any, Callable, Dict, List, MutableMapping, Optional, Tuple

import numpy as np
import pandas as pd
from IPython.display import display
from pydantic import BaseModel
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from vision_agent.tools.tools_types import BoundingBoxes
from vision_agent.utils.exceptions import RemoteToolCallFailed
from vision_agent.utils.execute import Error, MimeType
from vision_agent.utils.image_utils import normalize_bbox
from vision_agent.utils.type_defs import LandingaiAPIKey

_LOGGER = logging.getLogger(__name__)
_LND_API_KEY = os.environ.get("LANDINGAI_API_KEY", LandingaiAPIKey().api_key)
_LND_BASE_URL = os.environ.get("LANDINGAI_URL", "https://api.landing.ai")
_LND_API_URL = f"{_LND_BASE_URL}/v1/agent/model"
_LND_API_URL_v2 = f"{_LND_BASE_URL}/v1/tools"


def should_report_tool_traces() -> bool:
    return bool(os.environ.get("REPORT_TOOL_TRACES", False))


class ToolCallTrace(BaseModel):
    endpoint_url: str
    type: str
    request: MutableMapping[str, Any]
    response: MutableMapping[str, Any]
    error: Optional[Error]
    files: Optional[List[tuple[str, str]]]


def send_inference_request(
    payload: Dict[str, Any],
    endpoint_name: str,
    files: Optional[List[Tuple[Any, ...]]] = None,
    v2: bool = False,
    metadata_payload: Optional[Dict[str, Any]] = None,
    is_form: bool = False,
) -> Any:
    url = f"{_LND_API_URL_v2 if v2 else _LND_API_URL}/{endpoint_name}"
    if "TOOL_ENDPOINT_URL" in os.environ:
        url = os.environ["TOOL_ENDPOINT_URL"]

    headers = {"apikey": _LND_API_KEY}
    if "TOOL_ENDPOINT_AUTH" in os.environ:
        headers["Authorization"] = os.environ["TOOL_ENDPOINT_AUTH"]
        headers.pop("apikey")

    if runtime_tag := os.environ.get("RUNTIME_TAG", ""):
        headers["runtime_tag"] = runtime_tag

    session = _create_requests_session(
        url=url,
        num_retry=3,
        headers=headers,
    )

    function_name = "unknown"
    if "function_name" in payload:
        function_name = payload["function_name"]
    elif metadata_payload is not None and "function_name" in metadata_payload:
        function_name = metadata_payload["function_name"]

    response = _call_post(url, payload, session, files, function_name, is_form)

    return response["data"]


def send_task_inference_request(
    payload: Dict[str, Any],
    task_name: str,
    files: Optional[List[Tuple[Any, ...]]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    is_form: bool = False,
) -> Any:
    url = f"{_LND_API_URL_v2}/{task_name}"
    headers = {"apikey": _LND_API_KEY}
    session = _create_requests_session(
        url=url,
        num_retry=3,
        headers=headers,
    )

    function_name = "unknown"
    if metadata is not None and "function_name" in metadata:
        function_name = metadata["function_name"]
    response = _call_post(url, payload, session, files, function_name, is_form)
    return response["data"]


def _create_requests_session(
    url: str, num_retry: int, headers: Dict[str, str]
) -> Session:
    """Create a requests session with retry"""
    session = Session()
    retries = Retry(
        total=num_retry,
        backoff_factor=2,
        raise_on_redirect=True,
        raise_on_status=False,
        allowed_methods=["GET", "POST", "PUT"],
        status_forcelist=[
            408,  # Request Timeout
            429,  # Too Many Requests (ie. rate limiter).
            502,  # Bad Gateway
            503,  # Service Unavailable (include cloud circuit breaker)
            504,  # Gateway Timeout
        ],
    )
    session.mount(url, HTTPAdapter(max_retries=retries if num_retry > 0 else 0))
    session.headers.update(headers)
    return session


def get_tool_documentation(funcs: List[Callable[..., Any]]) -> str:
    docstrings = ""
    for func in funcs:
        docstrings += f"{func.__name__}{inspect.signature(func)}:\n{func.__doc__}\n\n"

    return docstrings


def get_tool_descriptions(funcs: List[Callable[..., Any]]) -> str:
    descriptions = ""
    for func in funcs:
        description = func.__doc__
        if description is None:
            description = ""

        if "Parameters:" in description:
            description = (
                description[: description.find("Parameters:")]
                .replace("\n", " ")
                .strip()
            )

        description = " ".join(description.split())
        descriptions += f"- {func.__name__}{inspect.signature(func)}: {description}\n"
    return descriptions


def get_tool_descriptions_by_names(
    tool_name: Optional[List[str]],
    funcs: List[Callable[..., Any]],
    util_funcs: List[
        Callable[..., Any]
    ],  # util_funcs will always be added to the list of functions
) -> str:
    if tool_name is None:
        return get_tool_descriptions(funcs + util_funcs)

    invalid_names = [
        name for name in tool_name if name not in {func.__name__ for func in funcs}
    ]

    if invalid_names:
        raise ValueError(f"Invalid customized tool names: {', '.join(invalid_names)}")

    filtered_funcs = (
        funcs
        if not tool_name
        else [func for func in funcs if func.__name__ in tool_name]
    )
    return get_tool_descriptions(filtered_funcs + util_funcs)


def get_tools_df(funcs: List[Callable[..., Any]]) -> pd.DataFrame:
    data: Dict[str, List[str]] = {"desc": [], "doc": [], "name": []}

    for func in funcs:
        desc = func.__doc__
        if desc is None:
            desc = ""
        desc = desc[: desc.find("Parameters:")].replace("\n", " ").strip()
        desc = " ".join(desc.split())

        doc = f"{func.__name__}{inspect.signature(func)}:\n{func.__doc__}"
        data["desc"].append(desc)
        data["doc"].append(doc)
        data["name"].append(func.__name__)

    return pd.DataFrame(data)  # type: ignore


def get_tools_info(funcs: List[Callable[..., Any]]) -> Dict[str, str]:
    data: Dict[str, str] = {}

    for func in funcs:
        desc = func.__doc__
        if desc is None:
            desc = ""

        data[func.__name__] = f"{func.__name__}{inspect.signature(func)}:\n{desc}"

    return data


def _call_post(
    url: str,
    payload: dict[str, Any],
    session: Session,
    files: Optional[List[Tuple[Any, ...]]] = None,
    function_name: str = "unknown",
    is_form: bool = False,
) -> Any:
    files_in_b64 = None
    if files:
        files_in_b64 = [(file[0], b64encode(file[1]).decode("utf-8")) for file in files]

    tool_call_trace = None
    try:
        if files is not None:
            response = session.post(url, data=payload, files=files)
        elif is_form:
            response = session.post(url, data=payload)
        else:
            response = session.post(url, json=payload)

        tool_call_trace_payload = (
            payload
            if "function_name" in payload
            else {**payload, **{"function_name": function_name}}
        )
        tool_call_trace = ToolCallTrace(
            endpoint_url=url,
            type="tool_call",
            request=tool_call_trace_payload,
            response={},
            error=None,
            files=files_in_b64,
        )

        if response.status_code != 200:
            tool_call_trace.error = Error(
                name="RemoteToolCallFailed",
                value=f"{response.status_code} - {response.text}",
                traceback_raw=[],
            )
            _LOGGER.error(f"Request failed: {response.status_code} {response.text}")
            raise RemoteToolCallFailed(
                function_name, response.status_code, response.text
            )

        result = response.json()
        tool_call_trace.response = result
        return result
    finally:
        if tool_call_trace is not None and should_report_tool_traces():
            trace = tool_call_trace.model_dump()
            display({MimeType.APPLICATION_JSON: trace}, raw=True)


def filter_bboxes_by_threshold(
    bboxes: BoundingBoxes, threshold: float
) -> BoundingBoxes:
    return list(filter(lambda bbox: bbox.score >= threshold, bboxes))


def add_bboxes_from_masks(
    all_preds: List[List[Dict[str, Any]]],
) -> List[List[Dict[str, Any]]]:
    for frame_preds in all_preds:
        for preds in frame_preds:
            if np.sum(preds["mask"]) == 0:
                preds["bbox"] = []
            else:
                rows, cols = np.where(preds["mask"])
                bbox = [
                    float(np.min(cols)),
                    float(np.min(rows)),
                    float(np.max(cols)),
                    float(np.max(rows)),
                ]
                bbox = normalize_bbox(bbox, preds["mask"].shape)
                preds["bbox"] = bbox

    return all_preds


def calculate_iou(bbox1: List[float], bbox2: List[float]) -> float:
    x1, y1, x2, y2 = bbox1
    x3, y3, x4, y4 = bbox2

    x_overlap = max(0, min(x2, x4) - max(x1, x3))
    y_overlap = max(0, min(y2, y4) - max(y1, y3))
    intersection = x_overlap * y_overlap

    area1 = (x2 - x1) * (y2 - y1)
    area2 = (x4 - x3) * (y4 - y3)
    union = area1 + area2 - intersection

    return intersection / union if union > 0 else 0


def single_nms(
    preds: List[Dict[str, Any]], iou_threshold: float
) -> List[Dict[str, Any]]:
    for i in range(len(preds)):
        for j in range(i + 1, len(preds)):
            if calculate_iou(preds[i]["bbox"], preds[j]["bbox"]) > iou_threshold:
                if preds[i]["score"] > preds[j]["score"]:
                    preds[j]["score"] = 0
                else:
                    preds[i]["score"] = 0

    return [pred for pred in preds if pred["score"] > 0]


def nms(
    all_preds: List[List[Dict[str, Any]]], iou_threshold: float
) -> List[List[Dict[str, Any]]]:
    if not isinstance(all_preds[0], List):
        all_preds = [all_preds]

    return_preds = []
    for frame_preds in all_preds:
        frame_preds = single_nms(frame_preds, iou_threshold)
        return_preds.append(frame_preds)

    return return_preds



from pydantic import Field, field_validator
from pydantic_settings import BaseSettings

from vision_agent.utils.exceptions import InvalidApiKeyError


class LandingaiAPIKey(BaseSettings):
    """The API key of a user in a particular organization in LandingLens.
    It supports loading from environment variables or .env files.
    The supported name of the environment variables are (case-insensitive):
    - LANDINGAI_API_KEY

    Environment variables will always take priority over values loaded from a dotenv file.
    """

    api_key: str = Field(
        default="land_sk_zKvyPcPV2bVoq7q87KwduoerAxuQpx33DnqP8M1BliOCiZOSoI",
        alias="LANDINGAI_API_KEY",
        description="The API key of LandingAI.",
    )

    @field_validator("api_key")
    @classmethod
    def is_api_key_valid(cls, key: str) -> str:
        """Check if the API key is a v2 key."""
        if not key:
            raise InvalidApiKeyError(f"LandingAI API key is required, but it's {key}")
        if not key.startswith("land_sk_"):
            raise InvalidApiKeyError(
                f"LandingAI API key (v2) must start with 'land_sk_' prefix, but it's {key}. See https://support.landing.ai/docs/api-key for more information."
            )
        return key

    class Config:
        env_file = ".env"
        env_prefix = "landingai_"
        case_sensitive = False
        extra = "ignore"



import copy
import json
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

from vision_agent.agent import Agent, AgentCoder, VisionAgentCoderV2
from vision_agent.agent.agent_utils import (
    add_media_to_chat,
    convert_message_to_agentmessage,
    extract_tag,
    format_conversation,
)
from vision_agent.agent.types import (
    AgentMessage,
    CodeContext,
    InteractionContext,
    PlanContext,
)
from vision_agent.agent.vision_agent_coder_v2 import format_code_context
from vision_agent.agent.vision_agent_prompts_v2 import CONVERSATION
from vision_agent.configs import Config
from vision_agent.lmm import LMM
from vision_agent.lmm.types import Message
from vision_agent.utils.execute import CodeInterpreter, CodeInterpreterFactory

CONFIG = Config()


def extract_conversation(
    chat: List[AgentMessage], include_conv: bool = False
) -> Tuple[List[AgentMessage], Optional[str]]:
    chat = copy.deepcopy(chat)

    # if we are in the middle of an interaction, return all the intermediate planning
    # steps
    if check_for_interaction(chat):
        return chat, None

    extracted_chat = []
    for chat_i in chat:
        if chat_i.role == "user":
            extracted_chat.append(chat_i)
        elif chat_i.role == "coder":
            if "<final_code>" in chat_i.content:
                extracted_chat.append(chat_i)
        elif include_conv and chat_i.role == "conversation":
            extracted_chat.append(chat_i)

    # only keep the last <final_code> and <final_test>
    final_code = None
    extracted_chat_strip_code: List[AgentMessage] = []
    for chat_i in reversed(extracted_chat):
        if "<final_code>" in chat_i.content and final_code is None:
            extracted_chat_strip_code = [chat_i] + extracted_chat_strip_code
            final_code = extract_tag(chat_i.content, "final_code")
            if final_code is not None:
                test_code = extract_tag(chat_i.content, "final_test")
                final_code += "\n" + test_code if test_code is not None else ""

        if "<final_code>" in chat_i.content and final_code is not None:
            continue

        extracted_chat_strip_code = [chat_i] + extracted_chat_strip_code

    return extracted_chat_strip_code, final_code


def run_conversation(agent: LMM, chat: List[AgentMessage]) -> str:
    extracted_chat, _ = extract_conversation(chat, include_conv=True)

    conv = format_conversation(extracted_chat)
    prompt = CONVERSATION.format(
        conversation=conv,
    )
    response = agent([{"role": "user", "content": prompt}], stream=False)
    return cast(str, response)


def check_for_interaction(chat: List[AgentMessage]) -> bool:
    return (
        len(chat) > 2
        and chat[-2].role == "interaction"
        and chat[-1].role == "interaction_response"
    )


def maybe_run_action(
    coder: AgentCoder,
    action: Optional[str],
    chat: List[AgentMessage],
    code_interpreter: Optional[CodeInterpreter] = None,
) -> Optional[List[AgentMessage]]:
    extracted_chat, final_code = extract_conversation(chat)
    if action == "generate_or_edit_vision_code":
        # there's an issue here because coder.generate_code will send it's code_context
        # to the outside user via it's update_callback, but we don't necessarily have
        # access to that update_callback here, so we re-create the message using
        # format_code_context.
        context = coder.generate_code(extracted_chat, code_interpreter=code_interpreter)

        if isinstance(context, CodeContext):
            return [
                AgentMessage(role="coder", content=format_code_context(context)),
                AgentMessage(role="observation", content=context.test_result.text()),
            ]
        elif isinstance(context, InteractionContext):
            return [
                AgentMessage(
                    role="interaction",
                    content=json.dumps([elt.model_dump() for elt in context.chat]),
                )
            ]
    elif action == "edit_code":
        # We don't want to pass code in plan_context.code so the coder will generate
        # new code from plan_context.plan
        plan_context = PlanContext(
            plan="Edit the latest code observed in the fewest steps possible according to the user's feedback."
            + ("<code>\n" + final_code + "\n</code>" if final_code is not None else ""),
            instructions=[
                chat_i.content
                for chat_i in extracted_chat
                if chat_i.role == "user" and "<final_code>" not in chat_i.content
            ],
            code="",
        )

        context = coder.generate_code_from_plan(
            extracted_chat, plan_context, code_interpreter=code_interpreter
        )
        return [
            AgentMessage(role="coder", content=format_code_context(context)),
            AgentMessage(role="observation", content=context.test_result.text()),
        ]
    elif action == "view_image":
        pass

    return None


class VisionAgentV2(Agent):
    """VisionAgentV2 is a conversational agent that allows you to more easily use a
    coder agent such as VisionAgentCoderV2 to write vision code for you.
    """

    def __init__(
        self,
        agent: Optional[LMM] = None,
        coder: Optional[AgentCoder] = None,
        hil: bool = False,
        verbose: bool = False,
        code_sandbox_runtime: Optional[str] = None,
        update_callback: Callable[[Dict[str, Any]], None] = lambda x: None,
    ) -> None:
        """Initialize the VisionAgentV2.

        Parameters:
            agent (Optional[LMM]): The language model to use for the agent. If None, a
                default AnthropicLMM will be used.
            coder (Optional[AgentCoder]): The coder agent to use for generating vision
                code. If None, a default VisionAgentCoderV2 will be used.
            hil (bool): Whether to use human-in-the-loop mode.
            verbose (bool): Whether to print out debug information.
            code_sandbox_runtime (Optional[str]): The code sandbox runtime to use, can
                be one of: None, "local" or "e2b". If None, it will read from the
                environment variable CODE_SANDBOX_RUNTIME.
            update_callback (Callable[[Dict[str, Any]], None]): The callback function
                that will send back intermediate conversation messages.
        """

        self.agent = agent if agent is not None else CONFIG.create_agent()
        self.coder = (
            coder
            if coder is not None
            else VisionAgentCoderV2(
                verbose=verbose, update_callback=update_callback, hil=hil
            )
        )

        self.verbose = verbose
        self.code_sandbox_runtime = code_sandbox_runtime
        self.update_callback = update_callback

        # force coder to use the same update_callback
        if hasattr(self.coder, "update_callback"):
            self.coder.update_callback = update_callback

    def __call__(
        self,
        input: Union[str, List[Message]],
        media: Optional[Union[str, Path]] = None,
    ) -> str:
        """Conversational interface to the agent. This is the main method to use to
        interact with the agent. It takes in a string or list of messages and returns
        the agent's response as a string.

        Parameters:
            input (Union[str, List[Message]]): The input to the agent. This can be a
                string or a list of messages in the format of [{"role": "user",
                "content": "describe your task here..."}, ...].
            media (Optional[Union[str, Path]]): The path to the media file to use with
                the input. This can be an image or video file.

        Returns:
            str: The agent's response as a string.
        """

        input_msg = convert_message_to_agentmessage(input, media)
        return self.chat(input_msg)[-1].content

    def chat(
        self,
        chat: List[AgentMessage],
        code_interpreter: Optional[CodeInterpreter] = None,
    ) -> List[AgentMessage]:
        """Conversational interface to the agent. This is the main method to use to
        interact with the agent. It takes in a list of messages and returns the agent's
        response as a list of messages.

        Parameters:
            chat (List[AgentMessage]): The input to the agent. This should be a list of
                AgentMessage objects.
            code_interpreter (Optional[CodeInterpreter]): The code interpreter to use.

        Returns:
            List[AgentMessage]: The agent's response as a list of AgentMessage objects.
        """

        chat = copy.deepcopy(chat)
        if not chat or chat[-1].role not in {"user", "interaction_response"}:
            raise ValueError(
                f"Last chat message must be from the user or interaction_response, got {chat[-1].role}."
            )

        return_chat = []
        with (
            CodeInterpreterFactory.new_instance(self.code_sandbox_runtime)
            if code_interpreter is None
            else code_interpreter
        ) as code_interpreter:
            int_chat, _, _ = add_media_to_chat(chat, code_interpreter)

            # if we had an interaction and then recieved an observation from the user
            # go back into the same action to finish it.
            action = None
            if check_for_interaction(int_chat):
                action = "generate_or_edit_vision_code"
            else:
                response_context = run_conversation(self.agent, int_chat)
                return_chat.append(
                    AgentMessage(role="conversation", content=response_context)
                )
                self.update_callback(return_chat[-1].model_dump())
                action = extract_tag(response_context, "action")

            updated_chat = maybe_run_action(
                self.coder, action, int_chat, code_interpreter=code_interpreter
            )

            # return an interaction early to get users feedback
            if updated_chat is not None and updated_chat[-1].role == "interaction":
                return_chat.extend(updated_chat)
            elif updated_chat is not None and updated_chat[-1].role != "interaction":
                # do not append updated_chat to return_chat becuase the observation
                # from running the action will have already been added via the callbacks
                obs_response_context = run_conversation(
                    self.agent, int_chat + return_chat + updated_chat
                )
                return_chat.append(
                    AgentMessage(role="conversation", content=obs_response_context)
                )
                self.update_callback(return_chat[-1].model_dump())

        return return_chat

    def log_progress(self, data: Dict[str, Any]) -> None:
        pass



from enum import Enum
from typing import List, Optional, Tuple, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, SerializationInfo, field_serializer


class BboxInput(BaseModel):
    image_path: str
    labels: List[str]
    bboxes: List[Tuple[int, int, int, int]]


class BboxInputBase64(BaseModel):
    image: str
    filename: str
    labels: List[str]
    bboxes: List[Tuple[int, int, int, int]]


class PromptTask(str, Enum):
    """Valid task prompts options for the Florence2 model."""

    PHRASE_GROUNDING = "<CAPTION_TO_PHRASE_GROUNDING>"


class Florence2FtRequest(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    image: Optional[str] = None
    video: Optional[bytes] = None
    task: PromptTask
    prompt: Optional[str] = ""
    chunk_length_frames: Optional[int] = None
    postprocessing: Optional[str] = None
    job_id: Optional[UUID] = Field(None, alias="jobId")

    @field_serializer("job_id")
    def serialize_job_id(self, job_id: UUID, _info: SerializationInfo) -> str:
        return str(job_id)


class JobStatus(str, Enum):
    """The status of a fine-tuning job.

    CREATED:
        The job has been created and is waiting to be scheduled to run.
    STARTING:
        The job has started running, but not entering the training phase.
    TRAINING:
        The job is training a model.
    EVALUATING:
        The job is evaluating the model and computing metrics.
    PUBLISHING:
        The job is exporting the artifact(s) to an external directory (s3 or local).
    SUCCEEDED:
        The job has finished, including training, evaluation and publishing the
        artifact(s).
    FAILED:
        The job has failed for some reason internally, it can be due to resources
        issues or the code itself.
    STOPPED:
        The job has been stopped by the use locally or in the cloud.
    """

    CREATED = "CREATED"
    STARTING = "STARTING"
    TRAINING = "TRAINING"
    EVALUATING = "EVALUATING"
    PUBLISHING = "PUBLISHING"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"
    STOPPED = "STOPPED"


class ODResponseData(BaseModel):
    label: str
    score: float
    bbox: Union[list[int], list[float]] = Field(alias="bounding_box")

    model_config = ConfigDict(
        populate_by_name=True,
    )


BoundingBoxes = list[ODResponseData]



import difflib
import json
import os
import re
import subprocess
import tempfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import libcst as cst
from IPython.display import display

import vision_agent as va
from vision_agent.clients.landing_public_api import LandingPublicAPI
from vision_agent.lmm.types import Message
from vision_agent.tools.tool_utils import get_tool_documentation
from vision_agent.tools.tools import TOOL_DESCRIPTIONS
from vision_agent.tools.tools_types import BboxInput, BboxInputBase64, PromptTask
from vision_agent.utils.execute import Execution, MimeType
from vision_agent.utils.image_utils import convert_to_b64

CURRENT_FILE = None
CURRENT_LINE = 0
DEFAULT_WINDOW_SIZE = 100
ZMQ_PORT = os.environ.get("ZMQ_PORT", None)


def report_progress_callback(port: int, inp: Dict[str, Any]) -> None:
    import zmq

    context = zmq.Context()
    socket = context.socket(zmq.PUSH)
    socket.connect(f"tcp://localhost:{port}")
    socket.send_json(inp)


def redisplay_results(execution: Execution) -> None:
    """This function is used to add previous execution results to the current output.
    This is handy if you are inside a notebook environment, call it notebook1, and you
    have a nested notebook environment, call it notebook2, and you want the execution
    results from notebook2 to be included in the execution results for notebook1.
    """
    for result in execution.results:
        if result.text is not None:
            display({MimeType.TEXT_PLAIN: result.text}, raw=True)
        if result.html is not None:
            display({MimeType.TEXT_HTML: result.html}, raw=True)
        if result.markdown is not None:
            display({MimeType.TEXT_MARKDOWN: result.markdown}, raw=True)
        if result.svg is not None:
            display({MimeType.IMAGE_SVG: result.svg}, raw=True)
        if result.png is not None:
            display({MimeType.IMAGE_PNG: result.png}, raw=True)
        if result.jpeg is not None:
            display({MimeType.IMAGE_JPEG: result.jpeg}, raw=True)
        if result.mp4 is not None:
            display({MimeType.VIDEO_MP4_B64: result.mp4}, raw=True)
        if result.latex is not None:
            display({MimeType.TEXT_LATEX: result.latex}, raw=True)
        if result.json is not None:
            display({MimeType.APPLICATION_JSON: result.json}, raw=True)
        if result.artifact is not None:
            display({MimeType.APPLICATION_ARTIFACT: result.artifact}, raw=True)
        if result.extra is not None:
            display(result.extra, raw=True)


class Artifacts:
    """Artifacts is a class that allows you to sync files between a local and remote
    environment. In our case, the remote environment could be where the VisionAgent is
    executing code and as the user adds new images, files or modifies files, those
    need to be in sync with the remote environment the VisionAgent is running in.
    """

    def __init__(self, cwd: Union[str, Path]) -> None:
        """Initializes the Artifacts object with it's remote and local save paths.

        Parameters:
            cwd (Union[str, Path]): The path to save all the chat related files. For example "/home/user/chat_abc/".
        """
        self.cwd = Path(cwd)

    def show(self) -> str:
        """Prints out all the files in the curret working directory"""
        output_str = "[Artifacts loaded]\n"
        for k in self:
            output_str += f"Artifact name: {k}, loaded to path: {str(self.cwd / k)}\n"
        output_str += "[End of artifacts]\n"
        print(output_str)
        return output_str

    def __iter__(self) -> Any:
        return iter(os.listdir(self.cwd))

    def __getitem__(self, name: str) -> Any:
        file_path = self.cwd / name
        if file_path.exists():
            with open(file_path, "r") as file:
                return file.read()
        else:
            raise KeyError(f"File '{name}' not found in artifacts")

    def __setitem__(self, name: str, value: Any) -> None:
        file_path = self.cwd / name
        with open(file_path, "w") as file:
            file.write(value)

    def __contains__(self, name: str) -> bool:
        return name in os.listdir(self.cwd)


def filter_file(file_name: Union[str, Path]) -> Tuple[bool, bool]:
    file_name_p = Path(file_name)
    return (
        file_name_p.is_file()
        and "__pycache__" not in str(file_name_p)
        and not file_name_p.name.startswith(".")
        and file_name_p.suffix
        in [".png", ".jpeg", ".jpg", ".mp4", ".txt", ".json", ".csv"]
    ), file_name_p.suffix in [".png", ".jpeg", ".jpg", ".mp4"]


# These tools are adapted from SWE-Agent https://github.com/princeton-nlp/SWE-agent


def format_lines(lines: List[str], start_idx: int) -> str:
    output = ""
    for i, line in enumerate(lines):
        output += f"{i + start_idx}|{line}"
    return output


def view_lines(
    lines: List[str],
    line_num: int,
    window_size: int,
    name: str,
    total_lines: int,
    print_output: bool = True,
) -> str:
    start = max(0, line_num - window_size)
    end = min(len(lines), line_num + window_size)
    return_str = (
        f"[Artifact: {name} ({total_lines} lines total)]\n"
        + format_lines(lines[start:end], start)
        + (
            "\n[End of artifact]"
            if end == len(lines)
            else f"\n[{len(lines) - end} more lines]"
        )
    )

    if print_output:
        print(return_str)
    return return_str


def open_code_artifact(
    artifacts: Artifacts, name: str, line_num: int = 0, window_size: int = 100
) -> str:
    """Opens the provided code artifact. If `line_num` is provided, the window will be
    moved to include that line. It only shows the first 100 lines by default! Max
    `window_size` supported is 2000.

    Parameters:
        artifacts (Artifacts): The artifacts object to open the artifact from.
        name (str): The name of the artifact to open.
        line_num (int): The line number to move the window to.
        window_size (int): The number of lines to show above and below the line.
    """
    if name not in artifacts:
        return f"[Artifact {name} does not exist]"

    total_lines = len(artifacts[name].splitlines())
    window_size = min(window_size, 2000)
    window_size = window_size // 2
    if line_num - window_size < 0:
        line_num = window_size
    elif line_num >= total_lines:
        line_num = total_lines - 1 - window_size

    lines = artifacts[name].splitlines(keepends=True)

    return view_lines(lines, line_num, window_size, name, total_lines)


def create_code_artifact(artifacts: Artifacts, name: str) -> str:
    """Creates a new code artifiact with the given name.

    Parameters:
        artifacts (Artifacts): The artifacts object to add the new artifact to.
        name (str): The name of the new artifact.
    """
    if name in artifacts:
        return_str = f"[Artifact {name} already exists]"
    else:
        artifacts[name] = ""
        return_str = f"[Artifact {name} created]"
    print(return_str)

    display(
        {
            MimeType.APPLICATION_ARTIFACT: json.dumps(
                {
                    "name": name,
                    "content": artifacts[name],
                    "action": "create",
                }
            )
        },
        raw=True,
    )
    return return_str


def edit_code_artifact(
    artifacts: Artifacts, name: str, start: int, end: int, content: str
) -> str:
    """Edits the given code artifact with the provided content. The content will be
    inserted between the `start` and `end` line numbers. If the `start` and `end` are
    the same, the content will be inserted at the `start` line number. If the `end` is
    greater than the total number of lines in the file, the content will be inserted at
    the end of the file. If the `start` or `end` are negative, the function will return
    an error message.

    Parameters:
        artifacts (Artifacts): The artifacts object to edit the artifact from.
        name (str): The name of the artifact to edit.
        start (int): The line number to start the edit, can be in [-1, total_lines]
            where -1 represents the end of the file.
        end (int): The line number to end the edit, can be in [-1, total_lines] where
            -1 represents the end of the file.
        content (str): The content to insert.
    """
    # just make the artifact if it doesn't exist instead of forcing agent to call
    # create_artifact
    if name not in artifacts:
        artifacts[name] = ""

    total_lines = len(artifacts[name].splitlines())
    if start == -1:
        start = total_lines
    if end == -1:
        end = total_lines

    if start < 0 or end < 0 or start > end or end > total_lines:
        print("[Invalid line range]")
        return "[Invalid line range]"

    new_content_lines = content.splitlines(keepends=True)
    new_content_lines = [
        line if line.endswith("\n") else line + "\n" for line in new_content_lines
    ]
    lines = artifacts[name].splitlines(keepends=True)
    lines = [line if line.endswith("\n") else line + "\n" for line in lines]
    edited_lines = lines[:start] + new_content_lines + lines[end:]

    cur_line = start + len(content.split("\n")) // 2
    with tempfile.NamedTemporaryFile(delete=True) as f:
        with open(f.name, "w") as f:  # type: ignore
            f.writelines(edited_lines)

        process = subprocess.Popen(
            [
                "flake8",
                "--isolated",
                "--select=F821,F822,F831,E111,E112,E113,E999,E902",
                f.name,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, _ = process.communicate()

        if stdout != "":
            stdout = stdout.replace(f.name, name)
            error_msg = "[Edit failed with the following status]\n" + stdout
            original_view = view_lines(
                lines,
                start + ((end - start) // 2),
                DEFAULT_WINDOW_SIZE,
                name,
                total_lines,
                print_output=False,
            )
            total_lines_edit = sum(1 for _ in edited_lines)
            edited_view = view_lines(
                edited_lines,
                cur_line,
                DEFAULT_WINDOW_SIZE,
                name,
                total_lines_edit,
                print_output=False,
            )

            error_msg += f"\n[This is how your edit would have looked like if applied]\n{edited_view}\n\n[This is the original code before your edit]\n{original_view}"
            print(error_msg)
            return error_msg

    artifacts[name] = "".join(edited_lines)

    display(
        {
            MimeType.APPLICATION_ARTIFACT: json.dumps(
                {
                    "name": name,
                    "content": artifacts[name],
                    "action": "edit",
                }
            )
        },
        raw=True,
    )
    return open_code_artifact(artifacts, name, cur_line)


def generate_vision_plan(
    artifacts: Artifacts,
    name: str,
    chat: str,
    media: List[str],
    test_multi_plan: bool = True,
    custom_tool_names: Optional[List[str]] = None,
) -> str:
    """Generates a plan to solve vision based tasks.

    Parameters:
        artifacts (Artifacts): The artifacts object to save the plan to.
        name (str): The name of the artifact to save the plan context to.
        chat (str): The chat message from the user.
        media (List[str]): The media files to use.
        test_multi_plan (bool): Do not change this parameter.
        custom_tool_names (Optional[List[str]]): Do not change this parameter.

    Returns:
        str: The generated plan.

    Examples
    --------
        >>> generate_vision_plan(artifacts, "plan.json", "Can you detect the dogs in this image?", ["image.jpg"])
        [Start Plan Context]
        plan1: This is a plan to detect dogs in an image
        -load image
        -detect dogs
        -return detections
        [End Plan Context]
    """

    # verbosity is set to 0 to avoid adding extra content to the VisionAgent conversation
    if ZMQ_PORT is not None:
        agent = va.agent.VisionAgentPlanner(
            report_progress_callback=lambda inp: report_progress_callback(
                int(ZMQ_PORT), inp
            ),
            verbosity=0,
        )
    else:
        agent = va.agent.VisionAgentPlanner(verbosity=0)

    fixed_chat: List[Message] = [{"role": "user", "content": chat, "media": media}]
    response = agent.generate_plan(
        fixed_chat,
        test_multi_plan=test_multi_plan,
        custom_tool_names=custom_tool_names,
    )
    if response.test_results is not None:
        redisplay_results(response.test_results)
    response.test_results = None
    artifacts[name] = response.model_dump_json()

    output_str = f"[Start Plan Context, saved at {name}]"
    for plan in response.plans.keys():
        output_str += f"\n{plan}: {response.plans[plan]['thoughts'].strip()}\n"  # type: ignore
        output_str += "    -" + "\n    -".join(
            e.strip() for e in response.plans[plan]["instructions"]
        )

    output_str += f"\nbest plan: {response.best_plan}\n"
    output_str += "thoughts: " + response.plan_thoughts.strip() + "\n"
    output_str += "[End Plan Context]"
    print(output_str)
    return output_str


def generate_vision_code(
    artifacts: Artifacts,
    name: str,
    chat: str,
    media: List[str],
    test_multi_plan: bool = True,
    custom_tool_names: Optional[List[str]] = None,
) -> str:
    """Generates python code to solve vision based tasks.

    Parameters:
        artifacts (Artifacts): The artifacts object to save the code to.
        name (str): The name of the artifact to save the code to.
        chat (str): The chat message from the user.
        media (List[str]): The media files to use.
        test_multi_plan (bool): Do not change this parameter.
        custom_tool_names (Optional[List[str]]): Do not change this parameter.

    Returns:
        str: The generated code.

    Examples
    --------
        >>> generate_vision_code(artifacts, "code.py", "Can you detect the dogs in this image?", ["image.jpg"])
        from vision_agent.tools import load_image, owl_v2
        def detect_dogs(image_path: str):
            image = load_image(image_path)
            dogs = owl_v2("dog", image)
            return dogs
    """
    # verbosity is set to 0 to avoid adding extra content to the VisionAgent conversation
    if ZMQ_PORT is not None:
        agent = va.agent.VisionAgentCoder(
            report_progress_callback=lambda inp: report_progress_callback(
                int(ZMQ_PORT), inp
            ),
            verbosity=0,
        )
    else:
        agent = va.agent.VisionAgentCoder(verbosity=0)

    fixed_chat: List[Message] = [{"role": "user", "content": chat, "media": media}]
    response = agent.generate_code(
        fixed_chat,
        test_multi_plan=test_multi_plan,
        custom_tool_names=custom_tool_names,
    )

    redisplay_results(response["test_result"])
    code = response["code"]
    artifacts[name] = code
    code_lines = code.splitlines(keepends=True)
    total_lines = len(code_lines)

    display(
        {
            MimeType.APPLICATION_ARTIFACT: json.dumps(
                {
                    "name": name,
                    "content": code,
                    "contentType": "vision_code",
                    "action": "create",
                }
            )
        },
        raw=True,
    )
    return view_lines(code_lines, 0, total_lines, name, total_lines)


def edit_vision_code(
    artifacts: Artifacts,
    name: str,
    chat_history: List[str],
    media: List[str],
    custom_tool_names: Optional[List[str]] = None,
) -> str:
    """Edits python code to solve a vision based task.

    Parameters:
        artifacts (Artifacts): The artifacts object to save the code to.
        name (str): The file path to the code.
        chat_history (List[str]): The chat history to used to generate the code.
        custom_tool_names (Optional[List[str]]): Do not change this parameter.

    Returns:
        str: The edited code.

    Examples
    --------
        >>> edit_vision_code(
        >>>     artifacts,
        >>>     "code.py",
        >>>     ["Can you detect the dogs in this image?", "Can you use a higher threshold?"],
        >>>     ["dog.jpg"],
        >>> )
        from vision_agent.tools import load_image, owl_v2
        def detect_dogs(image_path: str):
            image = load_image(image_path)
            dogs = owl_v2("dog", image, threshold=0.8)
            return dogs
    """

    # verbosity is set to 0 to avoid adding extra content to the VisionAgent conversation
    agent = va.agent.VisionAgentCoder(verbosity=0)
    if name not in artifacts:
        print(f"[Artifact {name} does not exist]")
        return f"[Artifact {name} does not exist]"

    code = artifacts[name]

    # Append latest code to second to last message from assistant
    fixed_chat_history: List[Message] = []
    user_message = "Previous user requests:"
    for i, chat in enumerate(chat_history):
        if i < len(chat_history) - 1:
            user_message += " " + chat
        else:
            fixed_chat_history.append(
                {"role": "user", "content": user_message, "media": media}
            )
            fixed_chat_history.append({"role": "assistant", "content": code})
            fixed_chat_history.append({"role": "user", "content": chat})

    response = agent.generate_code(
        fixed_chat_history,
        test_multi_plan=False,
        custom_tool_names=custom_tool_names,
    )

    redisplay_results(response["test_result"])
    code = response["code"]
    artifacts[name] = code
    code_lines = code.splitlines(keepends=True)
    total_lines = len(code_lines)

    display(
        {
            MimeType.APPLICATION_ARTIFACT: json.dumps(
                {
                    "name": name,
                    "content": code,
                    "action": "edit",
                }
            )
        },
        raw=True,
    )
    return view_lines(code_lines, 0, total_lines, name, total_lines)


def list_artifacts(artifacts: Artifacts) -> str:
    """Lists all the artifacts that have been loaded into the artifacts object."""
    output_str = artifacts.show()
    print(output_str)
    return output_str


def check_and_load_image(code: str) -> List[str]:
    if not code.strip():
        return []

    pattern = r"view_media_artifact\(\s*([^\)]+),\s*['\"]([^\)]+)['\"]\s*\)"
    matches = re.findall(pattern, code)
    return [match[1] for match in matches]


def view_media_artifact(artifacts: Artifacts, name: str) -> str:
    """Allows only the agent to view the media artifact with the given name. DO NOT use
    this to show media to the user, the user can already see all media saved in the
    artifacts.

    Parameters:
        artifacts (Artifacts): The artifacts object to show the image from.
        name (str): The name of the image artifact to show.
    """
    if name not in artifacts:
        output_str = f"[Artifact {name} does not exist]"
    else:
        output_str = f"[Image {name} displayed]"
    print(output_str)
    return output_str


def get_tool_descriptions() -> str:
    """Returns a description of all the tools that `generate_vision_code` has access to.
    Helpful for answering questions about what types of vision tasks you can do with
    `generate_vision_code`."""
    return TOOL_DESCRIPTIONS


def object_detection_fine_tuning(bboxes: List[Dict[str, Any]]) -> str:
    """DO NOT use this function unless the user has supplied you with bboxes.
    'object_detection_fine_tuning' is a tool that fine-tunes object detection models to
    be able to detect objects in an image based on a given dataset. It returns the fine
    tuning job id.

    Parameters:
        bboxes (List[BboxInput]): A list of BboxInput containing the image path, labels
            and bounding boxes. The coordinates are unnormalized.

    Returns:
        str: The fine tuning job id, this id will used to retrieve the fine tuned
            model.

    Example
    -------
        >>> fine_tuning_job_id = object_detection_fine_tuning(
            [{'image_path': 'filename.png', 'labels': ['screw'], 'bboxes': [[370, 30, 560, 290]]},
             {'image_path': 'filename.png', 'labels': ['screw'], 'bboxes': [[120, 0, 300, 170]]}],
             "phrase_grounding"
        )
    """
    task = "phrase_grounding"
    bboxes_input = [BboxInput.model_validate(bbox) for bbox in bboxes]
    task_type = PromptTask[task.upper()]
    fine_tuning_request = [
        BboxInputBase64(
            image=convert_to_b64(bbox_input.image_path),
            filename=Path(bbox_input.image_path).name,
            labels=bbox_input.labels,
            bboxes=bbox_input.bboxes,
        )
        for bbox_input in bboxes_input
    ]
    landing_api = LandingPublicAPI()
    fine_tune_id = str(
        landing_api.launch_fine_tuning_job("florencev2", task_type, fine_tuning_request)
    )
    print(f"[Fine tuning id: {fine_tune_id}]")
    return fine_tune_id


def get_diff(before: str, after: str) -> str:
    return "".join(
        difflib.unified_diff(
            before.splitlines(keepends=True), after.splitlines(keepends=True)
        )
    )


def get_diff_with_prompts(name: str, before: str, after: str) -> str:
    diff = get_diff(before, after)
    return f"[Artifact {name} edits]\n{diff}\n[End of edits]"


def use_extra_vision_agent_args(
    code: Optional[str],
    test_multi_plan: bool = True,
    custom_tool_names: Optional[List[str]] = None,
) -> Optional[str]:
    """This is for forcing arguments passed by the user to VisionAgent into the
    VisionAgentCoder call.

    Parameters:
        code (str): The code to edit.
        test_multi_plan (bool): Do not change this parameter.
        custom_tool_names (Optional[List[str]]): Do not change this parameter.

    Returns:
        str: The edited code.
    """
    if code is None:
        return None

    class VisionAgentTransformer(cst.CSTTransformer):
        def __init__(
            self, test_multi_plan: bool, custom_tool_names: Optional[List[str]]
        ):
            self.test_multi_plan = test_multi_plan
            self.custom_tool_names = custom_tool_names

        def leave_Call(
            self, original_node: cst.Call, updated_node: cst.Call
        ) -> cst.Call:
            # Check if the function being called is generate_vision_code or edit_vision_code
            if isinstance(updated_node.func, cst.Name) and updated_node.func.value in [
                "generate_vision_code",
                "edit_vision_code",
            ]:
                # Add test_multi_plan argument to generate_vision_code calls
                if updated_node.func.value == "generate_vision_code":
                    new_arg = cst.Arg(
                        keyword=cst.Name("test_multi_plan"),
                        value=cst.Name(str(self.test_multi_plan)),
                        equal=cst.AssignEqual(
                            whitespace_before=cst.SimpleWhitespace(""),
                            whitespace_after=cst.SimpleWhitespace(""),
                        ),
                    )
                    updated_node = updated_node.with_changes(
                        args=[*updated_node.args, new_arg]
                    )

                # Add custom_tool_names if provided
                if self.custom_tool_names is not None:
                    list_arg = []
                    for i, tool_name in enumerate(self.custom_tool_names):
                        if i < len(self.custom_tool_names) - 1:
                            list_arg.append(
                                cst._nodes.expression.Element(
                                    value=cst.SimpleString(value=f'"{tool_name}"'),
                                    comma=cst.Comma(
                                        whitespace_before=cst.SimpleWhitespace(""),
                                        whitespace_after=cst.SimpleWhitespace(" "),
                                    ),
                                )
                            )
                        else:
                            list_arg.append(
                                cst._nodes.expression.Element(
                                    value=cst.SimpleString(value=f'"{tool_name}"'),
                                )
                            )
                    new_arg = cst.Arg(
                        keyword=cst.Name("custom_tool_names"),
                        value=cst.List(list_arg),
                        equal=cst.AssignEqual(
                            whitespace_before=cst.SimpleWhitespace(""),
                            whitespace_after=cst.SimpleWhitespace(""),
                        ),
                    )
                    updated_node = updated_node.with_changes(
                        args=[*updated_node.args, new_arg]
                    )

            return updated_node

    # Parse the input code into a CST node
    tree = cst.parse_module(code)

    # Apply the transformer to modify the CST
    transformer = VisionAgentTransformer(test_multi_plan, custom_tool_names)
    modified_tree = tree.visit(transformer)

    # Return the modified code as a string
    return modified_tree.code


def use_object_detection_fine_tuning(
    artifacts: Artifacts, name: str, fine_tune_id: str
) -> str:
    """Replaces calls to 'owl_v2_image', 'florence2_phrase_detection' and
    'florence2_sam2_image' with the fine tuning id. This ensures that the code utilizes
    the fined tuned florence2 model. Returns the diff between the original code and the
    new code.

    Parameters:
        artifacts (Artifacts): The artifacts object to edit the code from.
        name (str): The name of the artifact to edit.
        fine_tune_id (str): The fine tuning job id.

    Examples
    --------
        >>> diff = use_object_detection_fine_tuning(artifacts, "code.py", "23b3b022-5ebf-4798-9373-20ef36429abf")
    """

    if name not in artifacts:
        output_str = f"[Artifact {name} does not exist]"
        print(output_str)
        return output_str

    code = artifacts[name]

    patterns_with_fine_tune_id = [
        (
            r'florence2_phrase_grounding\(\s*["\']([^"\']+)["\']\s*,\s*([^,]+)(?:,\s*["\'][^"\']+["\'])?\s*\)',
            lambda match: f'florence2_phrase_grounding("{match.group(1)}", {match.group(2)}, "{fine_tune_id}")',
        ),
        (
            r'florence2_phrase_grounding_video\(\s*["\']([^"\']+)["\']\s*,\s*([^,]+)(?:,\s*["\'][^"\']+["\'])?\s*\)',
            lambda match: f'florence2_phrase_grounding_video("{match.group(1)}", {match.group(2)}, "{fine_tune_id}")',
        ),
        (
            r'owl_v2_image\(\s*["\']([^"\']+)["\']\s*,\s*([^,]+)(?:,\s*["\'][^"\']+["\'])?\s*\)',
            lambda match: f'owl_v2_image("{match.group(1)}", {match.group(2)}, "{fine_tune_id}")',
        ),
        (
            r'florence2_sam2_image\(\s*["\']([^"\']+)["\']\s*,\s*([^,]+)(?:,\s*["\'][^"\']+["\'])?\s*\)',
            lambda match: f'florence2_sam2_image("{match.group(1)}", {match.group(2)}, "{fine_tune_id}")',
        ),
    ]

    new_code = code
    for (
        pattern_with_fine_tune_id,
        replacer_with_fine_tune_id,
    ) in patterns_with_fine_tune_id:
        if re.search(pattern_with_fine_tune_id, new_code):
            new_code = re.sub(
                pattern_with_fine_tune_id, replacer_with_fine_tune_id, new_code
            )

    if new_code == code:
        output_str = (
            f"[No function calls to replace with fine tuning id in artifact {name}]"
        )
        print(output_str)
        return output_str

    artifacts[name] = new_code

    diff = get_diff_with_prompts(name, code, new_code)
    print(diff)

    display(
        {
            MimeType.APPLICATION_ARTIFACT: json.dumps(
                {"name": name, "content": new_code, "action": "edit"}
            )
        },
        raw=True,
    )
    return diff


META_TOOL_DOCSTRING = get_tool_documentation(
    [
        get_tool_descriptions,
        open_code_artifact,
        create_code_artifact,
        edit_code_artifact,
        generate_vision_code,
        edit_vision_code,
        view_media_artifact,
        object_detection_fine_tuning,
        use_object_detection_fine_tuning,
        list_artifacts,
    ]
)



import os
from typing import List
from uuid import UUID

from requests.exceptions import HTTPError

from vision_agent.clients.http import BaseHTTP
from vision_agent.tools.tools_types import BboxInputBase64, JobStatus, PromptTask
from vision_agent.utils.exceptions import FineTuneModelNotFound
from vision_agent.utils.type_defs import LandingaiAPIKey


class LandingPublicAPI(BaseHTTP):
    def __init__(self) -> None:
        landing_url = os.environ.get("LANDINGAI_URL", "https://api.landing.ai")
        landing_api_key = os.environ.get("LANDINGAI_API_KEY", LandingaiAPIKey().api_key)
        headers = {"Content-Type": "application/json", "apikey": landing_api_key}
        super().__init__(base_endpoint=landing_url, headers=headers)

    def launch_fine_tuning_job(
        self, model_name: str, task: PromptTask, bboxes: List[BboxInputBase64]
    ) -> UUID:
        url = "v1/agent/jobs/fine-tuning"
        data = {
            "model": {"name": model_name, "task": task.value},
            "bboxes": [bbox.model_dump(by_alias=True) for bbox in bboxes],
        }
        response = self.post(url, payload=data)
        return UUID(response["jobId"])

    def check_fine_tuning_job(self, job_id: UUID) -> JobStatus:
        url = f"v1/agent/jobs/fine-tuning/{job_id}/status"
        try:
            get_job = self.get(url)
        except HTTPError as err:
            if err.response.status_code == 404:
                raise FineTuneModelNotFound()
        return JobStatus(get_job["status"])



SYSTEM_PROMPT = "You are a helpful assistant."

# EasyTool prompts
CHOOSE_PARAMS = (
    "This is an API tool documentation. Given a user's question, you need to output parameters according to the API tool documentation to successfully call the API to solve the user's question.\n"
    "This is the API tool documentation: {api_doc}\n"
    "Please note that: \n"
    "1. The Example in the API tool documentation can help you better understand the use of the API.\n"
    '2. Ensure the parameters you output are correct. The output must contain the required parameters, and can contain the optional parameters based on the question. If there are no parameters in the required parameters and optional parameters, just leave it as {{"Parameters":{{}}}}\n'
    "3. If the user's question mentions other APIs, you should ONLY consider the API tool documentation I give and do not consider other APIs.\n"
    '4. If you need to use this API multiple times, please set "Parameters" to a list.\n'
    "5. You must ONLY output in a parsible JSON format. Two example outputs look like:\n"
    "'''\n"
    'Example 1: {{"Parameters":{{"keyword": "Artificial Intelligence", "language": "English"}}}}\n'
    'Example 2: {{"Parameters":[{{"keyword": "Artificial Intelligence", "language": "English"}}, {{"keyword": "Machine Learning", "language": "English"}}]}}\n'
    "'''\n"
    "This is the user's question: {question}\n"
    "Output:\n"
)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from vision_agent.agent.types import (
    AgentMessage,
    CodeContext,
    InteractionContext,
    PlanContext,
)
from vision_agent.lmm.types import Message
from vision_agent.utils.execute import CodeInterpreter


class Agent(ABC):
    @abstractmethod
    def __call__(
        self,
        input: Union[str, List[Message]],
        media: Optional[Union[str, Path]] = None,
    ) -> Union[str, List[Message]]:
        pass

    @abstractmethod
    def log_progress(self, data: Dict[str, Any]) -> None:
        """Log the progress of the agent.
        This is a hook that is intended for reporting the progress of the agent.
        """
        pass


class AgentCoder(Agent):
    @abstractmethod
    def generate_code(
        self,
        chat: List[AgentMessage],
        max_steps: Optional[int] = None,
        code_interpreter: Optional[CodeInterpreter] = None,
    ) -> Union[CodeContext, InteractionContext]:
        pass

    @abstractmethod
    def generate_code_from_plan(
        self,
        chat: List[AgentMessage],
        plan_context: PlanContext,
        code_interpreter: Optional[CodeInterpreter] = None,
    ) -> CodeContext:
        pass


class AgentPlanner(Agent):
    @abstractmethod
    def generate_plan(
        self,
        chat: List[AgentMessage],
        max_steps: Optional[int] = None,
        code_interpreter: Optional[CodeInterpreter] = None,
    ) -> Union[PlanContext, InteractionContext]:
        pass



import json
import logging
from typing import Any, Dict, Optional

from requests import Session
from requests.adapters import HTTPAdapter

_LOGGER = logging.getLogger(__name__)


class BaseHTTP:
    _TIMEOUT = 30  # seconds
    _MAX_RETRIES = 3

    def __init__(
        self, base_endpoint: str, *, headers: Optional[Dict[str, Any]] = None
    ) -> None:
        self._headers = headers
        if headers is None:
            self._headers = {
                "Content-Type": "application/json",
            }
        self._base_endpoint = base_endpoint
        self._session = Session()
        self._session.headers.update(self._headers)  # type: ignore
        self._session.mount(
            self._base_endpoint, HTTPAdapter(max_retries=self._MAX_RETRIES)
        )

    def post(self, url: str, payload: Dict[str, Any]) -> Dict[str, Any]:
        formatted_url = f"{self._base_endpoint}/{url}"
        _LOGGER.info(f"Sending data to {formatted_url}")
        try:
            response = self._session.post(
                url=formatted_url, json=payload, timeout=self._TIMEOUT
            )
            response.raise_for_status()
            result: Dict[str, Any] = response.json()
            _LOGGER.info(json.dumps(result))
        except json.JSONDecodeError:
            resp_text = response.text
            _LOGGER.warning(f"Response seems incorrect: '{resp_text}'.")
            raise
        return result

    def get(self, url: str) -> Dict[str, Any]:
        formatted_url = f"{self._base_endpoint}/{url}"
        _LOGGER.info(f"Sending data to {formatted_url}")
        try:
            response = self._session.get(url=formatted_url, timeout=self._TIMEOUT)
            response.raise_for_status()
            result: Dict[str, Any] = response.json()
            _LOGGER.info(json.dumps(result))
        except json.JSONDecodeError:
            resp_text = response.text
            _LOGGER.warning(f"Response seems incorrect: '{resp_text}'.")
            raise
        return result



import copy
import json
import logging
import re
import sys
import tempfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import libcst as cst
from rich.console import Console
from rich.style import Style
from rich.syntax import Syntax
from rich.table import Table

import vision_agent.tools as T
from vision_agent.agent.types import AgentMessage, PlanContext
from vision_agent.lmm.types import Message
from vision_agent.utils.execute import CodeInterpreter, Execution
from vision_agent.utils.image_utils import b64_to_pil, convert_to_b64

logging.basicConfig(stream=sys.stdout)
_LOGGER = logging.getLogger(__name__)
_CONSOLE = Console()
_MAX_TABULATE_COL_WIDTH = 80


def _extract_sub_json(json_str: str) -> Optional[Dict[str, Any]]:
    json_pattern = r"\{.*\}"
    match = re.search(json_pattern, json_str, re.DOTALL)
    if match:
        json_str = match.group()
        try:
            # remove trailing comma
            trailing_bracket_pattern = r",\s+\}"
            json_str = re.sub(trailing_bracket_pattern, "}", json_str, flags=re.DOTALL)

            json_dict = json.loads(json_str)
            return json_dict  # type: ignore
        except json.JSONDecodeError:
            return None
    return None


def _find_markdown_json(json_str: str) -> str:
    pattern = r"```json(.*?)```"
    match = re.search(pattern, json_str, re.DOTALL)
    if match:
        return match.group(1).strip()
    return json_str


def _strip_markdown_code(inp_str: str) -> str:
    pattern = r"```python.*?```"
    cleaned_str = re.sub(pattern, "", inp_str, flags=re.DOTALL)
    return cleaned_str


def extract_json(json_str: str) -> Dict[str, Any]:
    json_str_mod = json_str.replace("\n", " ").strip()
    json_str_mod = json_str_mod.replace(": True", ": true").replace(
        ": False", ": false"
    )

    # sometimes the json is in single quotes
    try:
        return json.loads(json_str_mod.replace("'", '"'))  # type: ignore
    except json.JSONDecodeError:
        pass

    try:
        return json.loads(json_str_mod)  # type: ignore
    except json.JSONDecodeError:
        json_orig = json_str
        # don't replace quotes here or booleans since it can also introduce errors
        json_str = json_str.replace("\n", " ").strip()
        json_str = _strip_markdown_code(json_str)
        json_str = _find_markdown_json(json_str)
        json_dict = _extract_sub_json(json_str)

        if json_dict is None:
            error_msg = f"Could not extract JSON from the given str: {json_orig}"
            _LOGGER.exception(error_msg)
            raise json.JSONDecodeError(
                msg="Could not extract JSON", doc=json_orig, pos=0
            )

        return json_dict


def extract_code(code: str) -> str:
    if "\n```python" in code:
        start = "\n```python"
    elif "```python" in code:
        start = "```python"
    else:
        return code

    code = code[code.find(start) + len(start) :]
    code = code[: code.find("```")]
    if code.startswith("python\n"):
        code = code[len("python\n") :]
    return code


def extract_tag(
    content: str,
    tag: str,
) -> Optional[str]:
    inner_content = None
    remaning = content
    all_inner_content = []

    while f"<{tag}>" in remaning:
        inner_content_i = remaning[remaning.find(f"<{tag}>") + len(f"<{tag}>") :]
        if f"</{tag}>" not in inner_content_i:
            break
        inner_content_i = inner_content_i[: inner_content_i.find(f"</{tag}>")]
        remaning = remaning[remaning.find(f"</{tag}>") + len(f"</{tag}>") :]
        all_inner_content.append(inner_content_i)

    if len(all_inner_content) > 0:
        inner_content = "\n".join(all_inner_content)
    return inner_content


def remove_installs_from_code(code: str) -> str:
    pattern = r"\n!pip install.*?(\n|\Z)\n"
    code = re.sub(pattern, "", code, flags=re.DOTALL)
    return code


def format_feedback(memory: List[Dict[str, str]]) -> str:
    output_str = ""
    for i, m in enumerate(memory):
        output_str += f"### Feedback {i}:\n"
        output_str += f"Code {i}:\n```python\n{m['code']}```\n\n"
        output_str += f"Feedback {i}: {m['feedback']}\n\n"
        if "edits" in m:
            output_str += f"Edits {i}:\n{m['edits']}\n"
        output_str += "\n"

    return output_str


def format_plan_v2(plan: PlanContext) -> str:
    plan_str = plan.plan + "\n"
    plan_str += "Instructions:\n"
    for v in plan.instructions:
        plan_str += f"    - {v}\n"

    if plan.code:
        plan_str += "Code:\n"
        plan_str += plan.code
    return plan_str


def format_conversation(chat: List[AgentMessage]) -> str:
    chat = copy.deepcopy(chat)
    prompt = ""
    for chat_i in chat:
        if chat_i.role == "user" or chat_i.role == "coder":
            if "<final_code>" in chat_i.content:
                prompt += f"OBSERVATION: {chat_i.content}\n\n"
            elif chat_i.role == "user":
                prompt += f"USER: {chat_i.content}\n\n"
        elif chat_i.role == "conversation":
            prompt += f"AGENT: {chat_i.content}\n\n"
    return prompt


def format_plans(plans: Dict[str, Any]) -> str:
    plan_str = ""
    for k, v in plans.items():
        plan_str += "\n" + f"{k}: {v['thoughts']}\n"
        plan_str += "    -" + "\n    -".join([e for e in v["instructions"]])

    return plan_str


class DefaultImports:
    """Container for default imports used in the code execution."""

    common_imports = [
        "import os",
        "import numpy as np",
        "from vision_agent.tools import *",
        "from typing import *",
        "from pillow_heif import register_heif_opener",
        "register_heif_opener()",
    ]

    @staticmethod
    def to_code_string() -> str:
        return "\n".join(DefaultImports.common_imports + T.__new_tools__)

    @staticmethod
    def prepend_imports(code: str) -> str:
        """Run this method to prepend the default imports to the code.
        NOTE: be sure to run this method after the custom tools have been registered.
        """
        return DefaultImports.to_code_string() + "\n\n" + code


def print_code(title: str, code: str, test: Optional[str] = None) -> None:
    _CONSOLE.print(title, style=Style(bgcolor="dark_orange3", bold=True))
    _CONSOLE.print("=" * 30 + " Code " + "=" * 30)
    _CONSOLE.print(
        Syntax(
            code,
            "python",
            theme="gruvbox-dark",
            line_numbers=True,
            word_wrap=True,
        )
    )
    if test:
        _CONSOLE.print("=" * 30 + " Test " + "=" * 30)
        _CONSOLE.print(Syntax(test, "python", theme="gruvbox-dark", line_numbers=True))


def print_table(title: str, columns: List[str], rows: List[List[str]]) -> None:
    table = Table(title=title, show_header=True, header_style="bold magenta")
    for col in columns:
        table.add_column(col, style="cyan", no_wrap=True)

    for i, row in enumerate(rows):
        table.add_row(*row)
        if i < len(rows) - 1:
            table.add_row(*["-" * len(col) for col in row])
    _CONSOLE.print(table)


def add_media_to_chat(
    chat: List[AgentMessage], code_interpreter: Optional[CodeInterpreter] = None
) -> Tuple[List[AgentMessage], List[AgentMessage], List[Union[str, Path]]]:
    orig_chat = copy.deepcopy(chat)
    int_chat = copy.deepcopy(chat)
    media_list: List[Union[str, Path]] = []
    for chat_i in int_chat:
        if chat_i.media is not None:
            media_list_i: List[Union[str, Path]] = []
            for media in chat_i.media:
                if isinstance(media, str) and media.startswith("data:image/"):
                    media_pil = b64_to_pil(media)
                    with tempfile.NamedTemporaryFile(
                        mode="wb", suffix=".png", delete=False
                    ) as temp_file:
                        media_pil.save(temp_file, format="PNG")
                        media = str(temp_file.name)
                if code_interpreter is not None:
                    media = str(code_interpreter.upload_file(media))
                media_list_i.append(media)
                # don't duplicate appending media name and only add them for user messages
                if (
                    not str(chat_i.content).endswith(f" Media name {media}")
                    and chat_i.role == "user"
                ):
                    chat_i.content += f" Media name {media}"
            chat_i.media = media_list_i if len(media_list_i) > 0 else None
            media_list.extend(media_list_i)

    int_chat = cast(
        List[AgentMessage],
        [
            (
                AgentMessage(
                    role=c.role,
                    content=c.content,
                    media=c.media,
                )
                if c.media is not None
                else AgentMessage(role=c.role, content=c.content, media=None)
            )
            for c in int_chat
        ],
    )
    return int_chat, orig_chat, media_list


def capture_media_from_exec(execution: Execution) -> List[str]:
    images = []
    for result in execution.results:
        for format in result.formats():
            if format in ["png", "jpeg"]:
                # converts the image to png and then to base64
                images.append(
                    "data:image/png;base64,"
                    + convert_to_b64(b64_to_pil(result[format]))
                )
    return images


def convert_message_to_agentmessage(
    input: Union[str, List[Message]],
    media: Optional[Union[str, Path]] = None,
) -> List[AgentMessage]:
    if isinstance(input, str):
        input_msg = [
            AgentMessage(
                role="user",
                content=input,
                media=([media] if media is not None else None),
            )
        ]
    else:
        input_msg = [
            AgentMessage(role=msg["role"], content=msg["content"], media=None)
            for msg in input
        ]
        input_msg[0].media = [media] if media is not None else None
    return input_msg


def strip_function_calls(  # noqa: C901
    code: str, exclusions: Optional[List[str]] = None
) -> str:
    """This will strip out all code that calls functions except for functions included
    in exclusions.
    """
    if exclusions is None:
        exclusions = []

    def check_and_remove_node(node: cst.CSTNode, exclusions: List[str]) -> cst.CSTNode:
        if hasattr(node, "value") and isinstance(node.value, cst.Call):
            if (
                isinstance(node.value.func, cst.Name)
                and node.value.func.value in exclusions
            ):
                return node
            return cst.RemoveFromParent()  # type: ignore
        return node

    class StripFunctionCallsTransformer(cst.CSTTransformer):
        def __init__(self, exclusions: List[str]):
            # Store exclusions to skip removing certain function calls
            self.exclusions = exclusions
            self.in_function_or_class: List[bool] = []

        def visit_FunctionDef(self, node: cst.FunctionDef) -> Optional[bool]:
            self.in_function_or_class.append(True)
            return True

        def leave_FunctionDef(
            self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
        ) -> cst.BaseStatement:
            self.in_function_or_class.pop()
            return updated_node

        def visit_ClassDef(self, node: cst.ClassDef) -> Optional[bool]:
            self.in_function_or_class.append(True)
            return True

        def leave_ClassDef(
            self, node: cst.ClassDef, updated_node: cst.ClassDef
        ) -> cst.BaseStatement:
            self.in_function_or_class.pop()
            return updated_node

        def leave_Expr(
            self, original_node: cst.Expr, updated_node: cst.Expr
        ) -> cst.Expr:
            if not self.in_function_or_class:
                return cast(
                    cst.Expr, check_and_remove_node(updated_node, self.exclusions)
                )
            return updated_node

        def leave_Assign(
            self, original_node: cst.Assign, updated_node: cst.Assign
        ) -> cst.Assign:
            if not self.in_function_or_class:
                return cast(
                    cst.Assign, check_and_remove_node(updated_node, self.exclusions)
                )
            return updated_node

        def leave_If(self, original_node: cst.If, updated_node: cst.If) -> cst.If:
            if not self.in_function_or_class:
                return cast(
                    cst.If, check_and_remove_node(updated_node, self.exclusions)
                )
            return updated_node

        def leave_For(self, original_node: cst.For, updated_node: cst.For) -> cst.For:
            if not self.in_function_or_class:
                return cast(
                    cst.For, check_and_remove_node(updated_node, self.exclusions)
                )
            return updated_node

        def leave_While(
            self, original_node: cst.While, updated_node: cst.While
        ) -> cst.While:
            if not self.in_function_or_class:
                return cast(
                    cst.While, check_and_remove_node(updated_node, self.exclusions)
                )
            return updated_node

        def leave_With(
            self, original_node: cst.With, updated_node: cst.With
        ) -> cst.With:
            if not self.in_function_or_class:
                return cast(
                    cst.With, check_and_remove_node(updated_node, self.exclusions)
                )
            return updated_node

        def leave_Try(self, original_node: cst.Try, updated_node: cst.Try) -> cst.Try:
            if not self.in_function_or_class:
                return cast(
                    cst.Try, check_and_remove_node(updated_node, self.exclusions)
                )
            return updated_node

    tree = cst.parse_module(code)
    transformer = StripFunctionCallsTransformer(exclusions)
    modified_tree = tree.visit(transformer)
    return modified_tree.code



"""Vision Agent exceptions."""


class InvalidApiKeyError(Exception):
    """Exception raised when the an invalid API key is provided. This error could be raised from any SDK code, not limited to a HTTP client."""

    def __init__(self, message: str):
        self.message = f"""{message}
For more information, see https://landing-ai.github.io/landingai-python/landingai.html#manage-api-credentials"""
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class RemoteToolCallFailed(Exception):
    """Exception raised when an error occurs during a tool call."""

    def __init__(self, tool_name: str, status_code: int, message: str):
        self.message = (
            f"""Tool call ({tool_name}) failed due to {status_code} - {message}"""
        )


class RemoteSandboxError(Exception):
    """Exception related to remote sandbox."""

    is_retryable = False


class RemoteSandboxCreationError(RemoteSandboxError):
    """Exception raised when failed to create a remote sandbox.
    This could be due to the remote sandbox service is unavailable.
    """

    is_retryable = False


class RemoteSandboxExecutionError(RemoteSandboxError):
    """Exception raised when failed in a remote sandbox code execution."""

    is_retryable = False


class RemoteSandboxClosedError(RemoteSandboxError):
    """Exception raised when a remote sandbox is dead.
    This is retryable in the sense that the user can try again with a new sandbox. Can't be retried in the same sandbox.
    When this error is raised, the user should retry by create a new VisionAgent (i.e. a new sandbox).
    """

    is_retryable = True


class FineTuneModelIsNotReady(Exception):
    """Exception raised when the fine-tune model is not ready.
    If this is raised, it's recommended to wait 5 seconds before trying to use
    the model again.
    """


class FineTuneModelNotFound(Exception):
    """Exception raised when the fine-tune model is not found.
    If this is raised, it's recommended to try another model id.
    """



import copy
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Union, cast

from rich.console import Console
from rich.markup import escape

import vision_agent.tools.tools as T
from vision_agent.agent import AgentCoder, AgentPlanner
from vision_agent.agent.agent_utils import (
    DefaultImports,
    add_media_to_chat,
    capture_media_from_exec,
    convert_message_to_agentmessage,
    extract_tag,
    format_feedback,
    format_plan_v2,
    print_code,
    strip_function_calls,
)
from vision_agent.agent.types import (
    AgentMessage,
    CodeContext,
    InteractionContext,
    PlanContext,
)
from vision_agent.agent.vision_agent_coder_prompts_v2 import CODE, FIX_BUG, TEST
from vision_agent.agent.vision_agent_planner_v2 import VisionAgentPlannerV2
from vision_agent.configs import Config
from vision_agent.lmm import LMM
from vision_agent.lmm.types import Message
from vision_agent.tools.meta_tools import get_diff
from vision_agent.utils.execute import (
    CodeInterpreter,
    CodeInterpreterFactory,
    Execution,
)
from vision_agent.utils.sim import Sim, get_tool_recommender

CONFIG = Config()
_CONSOLE = Console()


def format_code_context(
    code_context: CodeContext,
) -> str:
    return f"<final_code>{code_context.code}</final_code>\n<final_test>{code_context.test}</final_test>"


def retrieve_tools(
    plan: List[str],
    tool_recommender: Sim,
) -> str:
    tool_docs = []
    for inst in plan:
        tools = tool_recommender.top_k(inst, k=1, thresh=0.3)
        tool_docs.extend([e["doc"] for e in tools])

    tool_docs_str = "\n\n".join(set(tool_docs))
    return tool_docs_str


def write_code(
    coder: LMM,
    chat: List[AgentMessage],
    tool_docs: str,
    plan: str,
) -> str:
    chat = copy.deepcopy(chat)
    if chat[-1].role != "user":
        raise ValueError("Last chat message must be from the user.")

    user_request = chat[-1].content
    prompt = CODE.format(
        docstring=tool_docs,
        question=user_request,
        plan=plan,
    )
    response = cast(str, coder([{"role": "user", "content": prompt}], stream=False))
    maybe_code = extract_tag(response, "code")

    # if the response wasn't properly formatted with the code tags just retrun the response
    if maybe_code is None:
        return response
    return maybe_code


def write_test(
    tester: LMM,
    chat: List[AgentMessage],
    tool_util_docs: str,
    code: str,
    media_list: Optional[Sequence[Union[str, Path]]] = None,
) -> str:
    chat = copy.deepcopy(chat)
    if chat[-1].role != "user":
        raise ValueError("Last chat message must be from the user.")

    user_request = chat[-1].content
    prompt = TEST.format(
        docstring=tool_util_docs,
        question=user_request,
        code=code,
        media=media_list,
    )
    response = cast(str, tester([{"role": "user", "content": prompt}], stream=False))
    maybe_code = extract_tag(response, "code")

    # if the response wasn't properly formatted with the code tags just retrun the response
    if maybe_code is None:
        return response
    return maybe_code


def debug_code(
    debugger: LMM,
    tool_docs: str,
    plan: str,
    code: str,
    test: str,
    result: Execution,
    debug_info: str,
    verbose: bool,
) -> tuple[str, str, str]:
    fixed_code = None
    fixed_test = None
    thoughts = ""
    success = False
    count = 0
    while not success and count < 3:
        try:
            # LLMs write worse code when it's in JSON, so we have it write JSON
            # followed by code each wrapped in markdown blocks.
            fixed_code_and_test_str = debugger(
                FIX_BUG.format(
                    docstring=tool_docs,
                    plan=plan,
                    code=code,
                    tests=test,
                    # Because of the way we trace function calls the trace information
                    # ends up in the results. We don't want to show this info to the
                    # LLM so we don't include it in the tool_output_str.
                    result="\n".join(
                        result.text(include_results=False).splitlines()[-50:]
                    ),
                    debug=debug_info,
                ),
                stream=False,
            )
            fixed_code_and_test_str = cast(str, fixed_code_and_test_str)
            thoughts_tag = extract_tag(fixed_code_and_test_str, "thoughts")
            thoughts = thoughts_tag if thoughts_tag is not None else ""
            fixed_code = extract_tag(fixed_code_and_test_str, "code")
            fixed_test = extract_tag(fixed_code_and_test_str, "test")

            success = not (fixed_code is None and fixed_test is None)

        except Exception as e:
            _CONSOLE.print(f"[bold red]Error while extracting JSON:[/bold red] {e}")

        count += 1

    old_code = code
    old_test = test

    if fixed_code is not None and fixed_code.strip() != "":
        code = fixed_code
    if fixed_test is not None and fixed_test.strip() != "":
        test = fixed_test

    debug_info_i = format_feedback(
        [
            {
                "code": f"{code}\n{test}",
                "feedback": thoughts,
                "edits": get_diff(f"{old_code}\n{old_test}", f"{code}\n{test}"),
            }
        ]
    )
    debug_info += f"\n{debug_info_i}"

    if verbose:
        _CONSOLE.print(
            f"[bold cyan]Thoughts on attempted fix:[/bold cyan] [green]{thoughts}[/green]"
        )

    return code, test, debug_info


def test_code(
    tester: LMM,
    debugger: LMM,
    chat: List[AgentMessage],
    plan: str,
    code: str,
    tool_docs: str,
    code_interpreter: CodeInterpreter,
    media_list: List[Union[str, Path]],
    verbose: bool,
) -> CodeContext:
    try:
        code = strip_function_calls(code)
    except Exception:
        # the code may be malformatted, this will fail in the exec call and the agent
        # will attempt to debug it
        pass
    test = write_test(
        tester=tester,
        chat=chat,
        tool_util_docs=T.UTILITIES_DOCSTRING,
        code=code,
        media_list=media_list,
    )
    if verbose:
        print_code("Code:", code)
        print_code("Test:", test)
    result = code_interpreter.exec_isolation(
        f"{DefaultImports.to_code_string()}\n{code}\n{test}"
    )
    if verbose:
        _CONSOLE.print(
            f"[bold cyan]Code execution result:[/bold cyan] [yellow]{escape(result.text(include_logs=True))}[/yellow]"
        )

    count = 0
    debug_info = ""
    while (not result.success or len(result.logs.stdout) == 0) and count < 3:
        code, test, debug_info = debug_code(
            debugger,
            T.UTILITIES_DOCSTRING + "\n" + tool_docs,
            plan,
            code,
            test,
            result,
            debug_info,
            verbose,
        )
        result = code_interpreter.exec_isolation(
            f"{DefaultImports.to_code_string()}\n{code}\n{test}"
        )
        count += 1
        if verbose:
            print_code("Code and test after attempted fix:", code, test)
            _CONSOLE.print(
                f"[bold cyan]Code execution result after attempted fix:[/bold cyan] [yellow]{escape(result.text(include_logs=True))}[/yellow]"
            )

    return CodeContext(
        code=f"{DefaultImports.to_code_string()}\n{code}",
        test=f"{DefaultImports.to_code_string()}\n{test}",
        success=result.success,
        test_result=result,
    )


def write_and_test_code(
    coder: LMM,
    tester: LMM,
    debugger: LMM,
    chat: List[AgentMessage],
    plan: str,
    tool_docs: str,
    code_interpreter: CodeInterpreter,
    media_list: List[Union[str, Path]],
    verbose: bool,
) -> CodeContext:
    code = write_code(
        coder=coder,
        chat=chat,
        tool_docs=tool_docs,
        plan=plan,
    )
    return test_code(
        tester,
        debugger,
        chat,
        plan,
        code,
        tool_docs,
        code_interpreter,
        media_list,
        verbose,
    )


class VisionAgentCoderV2(AgentCoder):
    """VisionAgentCoderV2 is an agent that will write vision code for you."""

    def __init__(
        self,
        planner: Optional[AgentPlanner] = None,
        coder: Optional[LMM] = None,
        tester: Optional[LMM] = None,
        debugger: Optional[LMM] = None,
        tool_recommender: Optional[Union[str, Sim]] = None,
        hil: bool = False,
        verbose: bool = False,
        code_sandbox_runtime: Optional[str] = None,
        update_callback: Callable[[Dict[str, Any]], None] = lambda _: None,
    ) -> None:
        """Initialize the VisionAgentCoderV2.

        Parameters:
            planner (Optional[AgentPlanner]): The planner agent to use for generating
                vision plans. If None, a default VisionAgentPlannerV2 will be used.
            coder (Optional[LMM]): The language model to use for the coder agent. If
                None, a default AnthropicLMM will be used.
            tester (Optional[LMM]): The language model to use for the tester agent. If
                None, a default AnthropicLMM will be used.
            debugger (Optional[LMM]): The language model to use for the debugger agent.
            tool_recommender (Optional[Union[str, Sim]]): The tool recommender to use.
            hil (bool): Whether to use human-in-the-loop mode.
            verbose (bool): Whether to print out debug information.
            code_sandbox_runtime (Optional[str]): The code sandbox runtime to use, can
                be one of: None, "local" or "e2b". If None, it will read from the
                environment variable CODE_SANDBOX_RUNTIME.
            update_callback (Callable[[Dict[str, Any]], None]): The callback function
                that will send back intermediate conversation messages.
        """

        self.planner = (
            planner
            if planner is not None
            else VisionAgentPlannerV2(
                verbose=verbose, update_callback=update_callback, hil=hil
            )
        )

        self.coder = coder if coder is not None else CONFIG.create_coder()
        self.tester = tester if tester is not None else CONFIG.create_tester()
        self.debugger = debugger if debugger is not None else CONFIG.create_debugger()
        if tool_recommender is not None:
            if isinstance(tool_recommender, str):
                self.tool_recommender = Sim.load(tool_recommender)
            elif isinstance(tool_recommender, Sim):
                self.tool_recommender = tool_recommender
        else:
            self.tool_recommender = get_tool_recommender()

        self.verbose = verbose
        self.code_sandbox_runtime = code_sandbox_runtime
        self.update_callback = update_callback
        if hasattr(self.planner, "update_callback"):
            self.planner.update_callback = update_callback

    def __call__(
        self,
        input: Union[str, List[Message]],
        media: Optional[Union[str, Path]] = None,
    ) -> str:
        """Generate vision code from a conversation.

        Parameters:
            input (Union[str, List[Message]]): The input to the agent. This can be a
                string or a list of messages in the format of [{"role": "user",
                "content": "describe your task here..."}, ...].
            media (Optional[Union[str, Path]]): The path to the media file to use with
                the input. This can be an image or video file.

        Returns:
            str: The generated code as a string.
        """

        input_msg = convert_message_to_agentmessage(input, media)
        code_or_interaction = self.generate_code(input_msg)
        if isinstance(code_or_interaction, InteractionContext):
            return code_or_interaction.chat[-1].content
        return code_or_interaction.code

    def generate_code(
        self,
        chat: List[AgentMessage],
        max_steps: Optional[int] = None,
        code_interpreter: Optional[CodeInterpreter] = None,
    ) -> Union[CodeContext, InteractionContext]:
        """Generate vision code from a conversation.

        Parameters:
            chat (List[AgentMessage]): The input to the agent. This should be a list of
                AgentMessage objects.
            code_interpreter (Optional[CodeInterpreter]): The code interpreter to use.

        Returns:
            CodeContext: The generated code as a CodeContext object which includes the
                code, test code, whether or not it was exceuted successfully, and the
                execution result.
        """

        chat = copy.deepcopy(chat)
        if not chat or chat[-1].role not in {"user", "interaction_response"}:
            raise ValueError(
                f"Last chat message must be from the user or interaction_response, got {chat[-1].role}."
            )

        with (
            CodeInterpreterFactory.new_instance(self.code_sandbox_runtime)
            if code_interpreter is None
            else code_interpreter
        ) as code_interpreter:
            int_chat, orig_chat, _ = add_media_to_chat(chat, code_interpreter)
            plan_context = self.planner.generate_plan(
                int_chat, max_steps=max_steps, code_interpreter=code_interpreter
            )
            # the planner needs an interaction, so return before generating code
            if isinstance(plan_context, InteractionContext):
                return plan_context

            code_context = self.generate_code_from_plan(
                orig_chat,
                plan_context,
                code_interpreter,
            )
        return code_context

    def generate_code_from_plan(
        self,
        chat: List[AgentMessage],
        plan_context: PlanContext,
        code_interpreter: Optional[CodeInterpreter] = None,
    ) -> CodeContext:
        """Generate vision code from a conversation and a previously made plan. This
        will skip the planning step and go straight to generating code.

        Parameters:
            chat (List[AgentMessage]): The input to the agent. This should be a list of
                AgentMessage objects.
            plan_context (PlanContext): The plan context that was previously generated.
                If plan_context.code is not provided, then the code will be generated
                from the chat messages.
            code_interpreter (Optional[CodeInterpreter]): The code interpreter to use.

        Returns:
            CodeContext: The generated code as a CodeContext object which includes the
                code, test code, whether or not it was exceuted successfully, and the
                execution result.
        """

        chat = copy.deepcopy(chat)
        if not chat or chat[-1].role not in {"user", "interaction_response"}:
            raise ValueError(
                f"Last chat message must be from the user or interaction_response, got {chat[-1].role}."
            )

        # we don't need the user_interaction response for generating code since it's
        # already in the plan context
        while len(chat) > 0 and chat[-1].role != "user":
            chat.pop()

        if not chat:
            raise ValueError("Chat must have at least one user message.")

        with (
            CodeInterpreterFactory.new_instance(self.code_sandbox_runtime)
            if code_interpreter is None
            else code_interpreter
        ) as code_interpreter:
            int_chat, _, media_list = add_media_to_chat(chat, code_interpreter)
            tool_docs = retrieve_tools(plan_context.instructions, self.tool_recommender)

            # If code is not provided from the plan_context then generate it, else use
            # the provided code and start with testing
            if not plan_context.code.strip():
                code = write_code(
                    coder=self.coder,
                    chat=int_chat,
                    tool_docs=tool_docs,
                    plan=format_plan_v2(plan_context),
                )
            else:
                code = plan_context.code

            code_context = test_code(
                tester=self.tester,
                debugger=self.debugger,
                chat=int_chat,
                plan=format_plan_v2(plan_context),
                code=code,
                tool_docs=tool_docs,
                code_interpreter=code_interpreter,
                media_list=media_list,
                verbose=self.verbose,
            )

        self.update_callback(
            {
                "role": "coder",
                "content": format_code_context(code_context),
                "media": capture_media_from_exec(code_context.test_result),
            }
        )
        self.update_callback(
            {
                "role": "observation",
                "content": code_context.test_result.text(),
            }
        )
        return code_context

    def log_progress(self, data: Dict[str, Any]) -> None:
        pass



import abc
import base64
import logging
import os
import platform
import re
import sys
import traceback
import warnings
from enum import Enum
from pathlib import Path
from time import sleep
from typing import Any, Dict, Iterable, List, Optional, Union

import nbformat
import tenacity
from dotenv import load_dotenv
from e2b.exceptions import SandboxException
from e2b_code_interpreter import CodeInterpreter as E2BCodeInterpreterImpl
from e2b_code_interpreter import Execution as E2BExecution
from e2b_code_interpreter import Result as E2BResult
from h11._util import LocalProtocolError
from httpx import ConnectError
from httpx import RemoteProtocolError as HttpcoreRemoteProtocolError
from httpx import RemoteProtocolError as HttpxRemoteProtocolError
from nbclient import NotebookClient
from nbclient import __version__ as nbclient_version
from nbclient.exceptions import CellTimeoutError, DeadKernelError
from nbclient.util import run_sync
from nbformat.v4 import new_code_cell
from opentelemetry.context import get_current
from opentelemetry.trace import SpanKind, Status, StatusCode, get_tracer
from pydantic import BaseModel, field_serializer
from typing_extensions import Self

from vision_agent.utils.exceptions import (
    RemoteSandboxCreationError,
    RemoteSandboxExecutionError,
)

load_dotenv()
_LOGGER = logging.getLogger(__name__)
_SESSION_TIMEOUT = 600  # 10 minutes
WORKSPACE = Path(os.getenv("WORKSPACE", ""))


class MimeType(str, Enum):
    """Represents a MIME type."""

    TEXT_PLAIN = "text/plain"
    TEXT_HTML = "text/html"
    TEXT_MARKDOWN = "text/markdown"
    IMAGE_SVG = "image/svg+xml"
    IMAGE_PNG = "image/png"
    IMAGE_JPEG = "image/jpeg"
    VIDEO_MP4_B64 = "video/mp4/base64"
    APPLICATION_PDF = "application/pdf"
    TEXT_LATEX = "text/latex"
    APPLICATION_JSON = "application/json"
    APPLICATION_JAVASCRIPT = "application/javascript"
    APPLICATION_ARTIFACT = "application/artifact"


class FileSerializer:
    """Adaptor class that allows IPython.display.display() to serialize a file to a
    base64 string representation.
    """

    def __init__(self, file_uri: str):
        self.video_uri = file_uri
        assert os.path.isfile(
            file_uri
        ), f"Only support local files currently: {file_uri}"
        assert Path(file_uri).exists(), f"File not found: {file_uri}"

    def __repr__(self) -> str:
        return f"FileSerializer({self.video_uri})"

    def base64(self) -> str:
        with open(self.video_uri, "rb") as file:
            return base64.b64encode(file.read()).decode("utf-8")


class Result:
    """Represents the data to be displayed as a result of executing a cell in a Jupyter
    notebook. The result is similar to the structure returned by ipython kernel:
    https://ipython.readthedocs.io/en/stable/development/execution.html#execution-semantics

    The result can contain multiple types of data, such as text, images, plots, etc.
    Each type of data is represented as a string, and the result can contain multiple
    types of data. The display calls don't have to have text representation, for the
    actual result the representation is always present for the result, the other
    representations are always optional.

    The class also provides methods to display the data in a Jupyter notebook.
    """

    text: Optional[str] = None
    html: Optional[str] = None
    markdown: Optional[str] = None
    svg: Optional[str] = None
    png: Optional[str] = None
    jpeg: Optional[str] = None
    pdf: Optional[str] = None
    mp4: Optional[str] = None
    latex: Optional[str] = None
    json: Optional[Dict[str, Any]] = None
    javascript: Optional[str] = None
    artifact: Optional[str] = None
    extra: Optional[Dict[str, Any]] = None
    "Extra data that can be included. Not part of the standard types."

    is_main_result: bool
    "Whether this data is the result of the cell. Data can be produced by display calls of which can be multiple in a cell."

    def __init__(self, is_main_result: bool, data: Dict[str, Any]):
        self.is_main_result = is_main_result
        self.text = data.pop(MimeType.TEXT_PLAIN, None)
        if self.text and (self.text.startswith("'") and self.text.endswith("'")):
            # This is a workaround for the issue that str result is wrapped with single quotes by notebook.
            # E.g. input text: "'flower'". what we want: "flower"
            self.text = self.text[1:-1]

        self.html = data.pop(MimeType.TEXT_HTML, None)
        self.markdown = data.pop(MimeType.TEXT_MARKDOWN, None)
        self.svg = data.pop(MimeType.IMAGE_SVG, None)
        self.png = data.pop(MimeType.IMAGE_PNG, None)
        self.jpeg = data.pop(MimeType.IMAGE_JPEG, None)
        self.pdf = data.pop(MimeType.APPLICATION_PDF, None)
        self.mp4 = data.pop(MimeType.VIDEO_MP4_B64, None)
        self.latex = data.pop(MimeType.TEXT_LATEX, None)
        self.json = data.pop(MimeType.APPLICATION_JSON, None)
        self.javascript = data.pop(MimeType.APPLICATION_JAVASCRIPT, None)
        self.artifact = data.pop(MimeType.APPLICATION_ARTIFACT, None)
        self.extra = data
        # Only keeping the PNG representation if both PNG and JPEG are present
        if self.png and self.jpeg:
            del self.jpeg

    # Allows to iterate over formats()
    def __getitem__(self, key: Any) -> Any:
        return getattr(self, key)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return str(self.text)

    def _repr_html_(self) -> Optional[str]:
        """Returns the HTML representation of the data."""
        return self.html

    def _repr_markdown_(self) -> Optional[str]:
        """Returns the Markdown representation of the data."""
        return self.markdown

    def _repr_svg_(self) -> Optional[str]:
        """Returns the SVG representation of the data."""
        return self.svg

    def _repr_png_(self) -> Optional[str]:
        """Returns the base64 representation of the PNG data."""
        return self.png

    def _repr_jpeg_(self) -> Optional[str]:
        """Returns the base64 representation of the JPEG data."""
        return self.jpeg

    def _repr_pdf_(self) -> Optional[str]:
        """Returns the PDF representation of the data."""
        return self.pdf

    def _repr_latex_(self) -> Optional[str]:
        """Returns the LaTeX representation of the data."""
        return self.latex

    def _repr_json_(self) -> Optional[dict]:
        """Returns the JSON representation of the data."""
        return self.json

    def _repr_javascript_(self) -> Optional[str]:
        """Returns the JavaScript representation of the data."""
        return self.javascript

    def formats(self) -> Iterable[str]:
        """Returns all available formats of the result.

        :return: All available formats of the result in MIME types.
        """
        formats = []
        if self.html:
            formats.append("html")
        if self.markdown:
            formats.append("markdown")
        if self.svg:
            formats.append("svg")
        if self.png:
            formats.append("png")
        if self.jpeg:
            formats.append("jpeg")
        if self.pdf:
            formats.append("pdf")
        if self.latex:
            formats.append("latex")
        if self.json:
            formats.append("json")
        if self.javascript:
            formats.append("javascript")
        if self.mp4:
            formats.append("mp4")
        if self.artifact:
            formats.append("artifact")
        if self.extra:
            formats.extend(iter(self.extra))
        return formats

    @staticmethod
    def from_e2b_result(result: E2BResult) -> "Result":
        """
        Creates a Result object from an E2BResult object.
        """
        data = {
            MimeType.TEXT_PLAIN.value: result.text,
            MimeType.IMAGE_PNG.value: result.png,
            MimeType.APPLICATION_JSON.value: result.json,
        }
        for k, v in result.extra.items():
            data[k] = v
        return Result(
            is_main_result=result.is_main_result,
            data=data,
        )


class Logs(BaseModel):
    """Data printed to stdout and stderr during execution, usually by print statements,
    logs, warnings, subprocesses, etc.
    """

    stdout: List[str] = []
    "List of strings printed to stdout by prints, subprocesses, etc."
    stderr: List[str] = []
    "List of strings printed to stderr by prints, subprocesses, etc."

    def __str__(self) -> str:
        stdout_str = "\n".join(self.stdout)
        stderr_str = "\n".join(self.stderr)
        return _remove_escape_and_color_codes(
            f"----- stdout -----\n{stdout_str}\n----- stderr -----\n{stderr_str}"
        )

    def to_json(self) -> dict[str, list[str]]:
        return {"stdout": self.stdout, "stderr": self.stderr}


class Error(BaseModel):
    """Represents an error that occurred during the execution of a cell. The error
    contains the name of the error, the value of the error, and the traceback.
    """

    name: str
    "Name of the exception."
    value: str
    "Value of the exception."
    traceback_raw: List[str]
    "List of strings representing the traceback."

    @property
    def traceback(self, return_clean_text: bool = True) -> str:
        """
        Returns the traceback as a single string.
        """
        text = "\n".join(self.traceback_raw)
        return _remove_escape_and_color_codes(text) if return_clean_text else text

    @staticmethod
    def from_exception(e: Exception) -> "Error":
        """
        Creates an Error object from an exception.
        """
        return Error(
            name=e.__class__.__name__,
            value=str(e),
            traceback_raw=traceback.format_exception(type(e), e, e.__traceback__),
        )


class Execution(BaseModel):
    """Represents the result of a cell execution."""

    class Config:
        arbitrary_types_allowed = True

    results: List[Result] = []
    "List of the result of the cell (interactively interpreted last line), display calls (e.g. matplotlib plots)."
    logs: Logs = Logs()
    "Logs printed to stdout and stderr during execution."
    error: Optional[Error] = None
    "Error object if an error occurred, None otherwise."

    def text(self, include_logs: bool = True, include_results: bool = True) -> str:
        """Returns the text representation of this object, i.e. including the main
        result or the error traceback, optionally along with the logs (stdout, stderr).
        """
        prefix = str(self.logs) if include_logs else ""
        if self.error:
            return prefix + "\n----- Error -----\n" + self.error.traceback

        if include_results:
            result_str = [
                (
                    f"----- Final output -----\n{res.text}"
                    if res.is_main_result
                    else f"----- Intermediate output-----\n{res.text}"
                )
                for res in self.results
            ]
            return prefix + "\n" + "\n".join(result_str)
        return prefix

    @property
    def success(self) -> bool:
        """
        Returns whether the execution was successful.
        """
        return self.error is None

    def get_main_result(self) -> Optional[Result]:
        """Get the main result of the execution. An execution may have multiple
        results, e.g. intermediate outputs. The main result is the last output of the
        cell execution.
        """
        if not self.success:
            _LOGGER.info("Result is not available as the execution was not successful.")
            return None
        if not self.results or not any(res.is_main_result for res in self.results):
            _LOGGER.info("Execution was successful but there is no main result.")
            return None
        main_result = self.results[-1]
        assert main_result.is_main_result, "The last result should be the main result."
        return main_result

    def to_json(self) -> str:
        """Returns the JSON representation of the Execution object."""
        return self.model_dump_json(exclude_none=True)

    @field_serializer("results", when_used="json")
    def serialize_results(results: List[Result]) -> List[Dict[str, Union[str, bool]]]:  # type: ignore
        """Serializes the results to JSON. This method is used by the Pydantic JSON
        encoder.
        """
        serialized = []
        for result in results:
            serialized_dict = {key: result[key] for key in result.formats()}

            serialized_dict["text"] = result.text
            serialized_dict["is_main_result"] = result.is_main_result
            serialized.append(serialized_dict)
        return serialized

    @staticmethod
    def from_exception(exec: Exception, traceback_raw: List[str]) -> "Execution":
        """Creates an Execution object from an exception."""
        return Execution(
            error=Error(
                name=exec.__class__.__name__,
                value=_remove_escape_and_color_codes(str(exec)),
                traceback_raw=[
                    _remove_escape_and_color_codes(line) for line in traceback_raw
                ],
            )
        )

    @staticmethod
    def from_e2b_execution(exec: E2BExecution) -> "Execution":
        """Creates an Execution object from an E2BResult object."""
        return Execution(
            results=[Result.from_e2b_result(res) for res in exec.results],
            logs=Logs(stdout=exec.logs.stdout, stderr=exec.logs.stderr),
            error=(
                Error(
                    name=exec.error.name,
                    value=_remove_escape_and_color_codes(exec.error.value),
                    traceback_raw=[
                        _remove_escape_and_color_codes(line)
                        for line in exec.error.traceback.split("\n")
                    ],
                )
                if exec.error
                else None
            ),
        )


class CodeInterpreter(abc.ABC):
    """Code interpreter interface."""

    def __init__(
        self,
        timeout: int,
        remote_path: Optional[Union[str, Path]] = None,
        non_exiting: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.timeout = timeout
        self.remote_path = Path(remote_path if remote_path is not None else WORKSPACE)
        self.non_exiting = non_exiting

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *exc_info: Any) -> None:
        if not self.non_exiting:
            self.close()

    def close(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError()

    def restart_kernel(self) -> None:
        raise NotImplementedError()

    def exec_cell(self, code: str) -> Execution:
        raise NotImplementedError()

    def exec_isolation(self, code: str) -> Execution:
        self.restart_kernel()
        return self.exec_cell(code)

    def upload_file(self, file: Union[str, Path]) -> Path:
        # Default behavior is a no-op (for local code interpreter)
        return Path(file)

    def download_file(
        self, remote_file_path: Union[str, Path], local_file_path: Union[str, Path]
    ) -> Path:
        # Default behavior is a no-op (for local code interpreter)
        return Path(local_file_path)


class E2BCodeInterpreter(CodeInterpreter):
    def __init__(
        self, remote_path: Optional[Union[str, Path]] = None, *args: Any, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        assert os.getenv("E2B_API_KEY"), "E2B_API_KEY environment variable must be set"
        try:
            self.interpreter = E2BCodeInterpreter._new_e2b_interpreter_impl(
                *args, **kwargs
            )
        except Exception as e:
            raise RemoteSandboxCreationError(
                f"Failed to create a remote sandbox due to {e}"
            ) from e

        result = self.exec_cell(
            """
import platform
import sys
import importlib.metadata

print(f"Python version: {sys.version}")
print(f"OS version: {platform.system()} {platform.release()} ({platform.architecture()})")
va_version = importlib.metadata.version("vision-agent")
print(f"Vision Agent version: {va_version}")"""
        )
        sys_versions = "\n".join(result.logs.stdout)
        _LOGGER.info(
            f"E2BCodeInterpreter (sandbox id: {self.interpreter.sandbox_id}) initialized:\n{sys_versions}"
        )
        self.remote_path = Path(
            remote_path if remote_path is not None else "/home/user"
        )

    def close(self, *args: Any, **kwargs: Any) -> None:
        try:
            self.interpreter.kill(request_timeout=2)
            _LOGGER.info(
                f"The sandbox {self.interpreter.sandbox_id} is closed successfully."
            )
        except Exception as e:
            _LOGGER.warn(
                f"Failed to close the remote sandbox ({self.interpreter.sandbox_id}) due to {e}. This is not an issue. It's likely that the sandbox is already closed due to timeout."
            )

    def restart_kernel(self) -> None:
        self.interpreter.notebook.restart_kernel()

    @tenacity.retry(
        wait=tenacity.wait_exponential_jitter(),
        stop=tenacity.stop_after_attempt(3),
        retry=tenacity.retry_if_exception_type(
            (
                LocalProtocolError,
                HttpxRemoteProtocolError,
                HttpcoreRemoteProtocolError,
                ConnectError,
                SandboxException,
            )
        ),
        before_sleep=tenacity.before_sleep_log(_LOGGER, logging.INFO),
        after=tenacity.after_log(_LOGGER, logging.INFO),
    )
    def exec_cell(self, code: str) -> Execution:
        self.interpreter.set_timeout(_SESSION_TIMEOUT)  # Extend the life of the sandbox
        try:
            _LOGGER.info(
                f"Start code execution in remote sandbox {self.interpreter.sandbox_id}. Timeout: {_SESSION_TIMEOUT}. Code hash: {hash(code)}"
            )
            execution = self.interpreter.notebook.exec_cell(
                code=code,
                on_stdout=lambda msg: _LOGGER.info(msg),
                on_stderr=lambda msg: _LOGGER.info(msg),
            )
            _LOGGER.info(
                f"Finished code execution in remote sandbox {self.interpreter.sandbox_id}. Code hash: {hash(code)}"
            )
            return Execution.from_e2b_execution(execution)
        except (
            LocalProtocolError,
            HttpxRemoteProtocolError,
            HttpcoreRemoteProtocolError,
            ConnectError,
            SandboxException,
        ) as e:
            raise e
        except Exception as e:
            raise RemoteSandboxExecutionError(
                f"Failed executing code in remote sandbox ({self.interpreter.sandbox_id}) due to error '{type(e).__name__} {str(e)}', code: {code}"
            ) from e

    @tenacity.retry(
        wait=tenacity.wait_exponential_jitter(),
        stop=tenacity.stop_after_attempt(3),
        retry=tenacity.retry_if_exception_type(
            (
                LocalProtocolError,
                HttpxRemoteProtocolError,
                HttpcoreRemoteProtocolError,
                ConnectError,
                SandboxException,
            )
        ),
        before_sleep=tenacity.before_sleep_log(_LOGGER, logging.INFO),
        after=tenacity.after_log(_LOGGER, logging.INFO),
    )
    def upload_file(self, file: Union[str, Path]) -> Path:
        file_name = Path(file).name
        with open(file, "rb") as f:
            self.interpreter.files.write(path=str(self.remote_path / file_name), data=f)
        _LOGGER.info(f"File ({file}) is uploaded to: {str(self.remote_path)}")
        return self.remote_path / file_name

    def download_file(
        self, remote_file_path: Union[str, Path], local_file_path: Union[str, Path]
    ) -> Path:
        with open(local_file_path, "w+b") as f:
            f.write(
                self.interpreter.files.read(path=str(remote_file_path), format="bytes")
            )
        _LOGGER.info(f"File ({remote_file_path}) is downloaded to: {local_file_path}")
        return Path(local_file_path)

    @staticmethod
    def _new_e2b_interpreter_impl(*args, **kwargs) -> E2BCodeInterpreterImpl:  # type: ignore
        template_name = os.environ.get("E2B_TEMPLATE_NAME", "va-sandbox")
        _LOGGER.info(
            f"Creating a new E2BCodeInterpreter using template: {template_name}"
        )
        return E2BCodeInterpreterImpl(template=template_name, *args, **kwargs)


class LocalCodeInterpreter(CodeInterpreter):
    def __init__(
        self,
        timeout: int = _SESSION_TIMEOUT,
        remote_path: Optional[Union[str, Path]] = None,
        non_exiting: bool = False,
    ) -> None:
        super().__init__(timeout=timeout, non_exiting=non_exiting)
        self.nb = nbformat.v4.new_notebook()
        # Set the notebook execution path to the remote path
        self.remote_path = Path(remote_path if remote_path is not None else WORKSPACE)
        self.resources = {"metadata": {"path": str(self.remote_path)}}
        self.nb_client = NotebookClient(
            self.nb,
            timeout=self.timeout,
            resources=self.resources,
        )
        _LOGGER.info(
            f"""Local code interpreter initialized
Python version: {sys.version}
OS version: {platform.system()} {platform.release()} ({platform.architecture()})
nbclient version: {nbclient_version}
nbformat version: {nbformat.__version__}
Timeout: {self.timeout}"""
        )
        sleep(1)
        self._new_kernel()

    def _new_kernel(self) -> None:
        if self.nb_client.kc is None or not run_sync(self.nb_client.kc.is_alive)():  # type: ignore
            self.nb_client.create_kernel_manager()
            self.nb_client.start_new_kernel()
            self.nb_client.start_new_kernel_client()

    def close(self) -> None:
        if self.nb_client.km is not None and run_sync(self.nb_client.km.is_alive)():  # type: ignore
            run_sync(self.nb_client.km.shutdown_kernel)(now=True)
            run_sync(self.nb_client.km.cleanup_resources)()

            if self.nb_client.kc is not None:
                channels = [
                    self.nb_client.kc.stdin_channel,
                    self.nb_client.kc.hb_channel,
                    self.nb_client.kc.control_channel,
                ]

                for ch in channels:
                    if ch.is_alive():
                        ch.stop()
                self.nb_client.kc.stop_channels()

                self.nb_client.kc = None
            self.nb_client.km = None

    def restart_kernel(self) -> None:
        self.close()
        self.nb = nbformat.v4.new_notebook()
        self.nb_client = NotebookClient(
            self.nb, timeout=self.timeout, resources=self.resources
        )
        sleep(1)
        self._new_kernel()

    def exec_cell(self, code: str) -> Execution:
        # track the exec_cell with opentelemetry trace
        tracer = get_tracer(__name__)
        context = get_current()
        with tracer.start_as_current_span(
            "notebook_cell_execution", kind=SpanKind.INTERNAL, context=context
        ) as span:
            try:
                # Add code as span attribute
                span.set_attribute("code", code)
                span.set_attribute("cell_index", len(self.nb.cells))

                self.nb.cells.append(new_code_cell(code))
                cell = self.nb.cells[-1]
                self.nb_client.execute_cell(cell, len(self.nb.cells) - 1)

                result = _parse_local_code_interpreter_outputs(
                    self.nb.cells[-1].outputs
                )
                span.set_status(Status(StatusCode.OK))
                return result
            except CellTimeoutError as e:
                run_sync(self.nb_client.km.interrupt_kernel)()  # type: ignore
                sleep(1)
                span.set_status(Status(StatusCode.ERROR, str(e)))
                span.record_exception(e)
                traceback_raw = traceback.format_exc().splitlines()
                return Execution.from_exception(e, traceback_raw)
            except DeadKernelError as e:
                self.restart_kernel()
                span.set_status(Status(StatusCode.ERROR, str(e)))
                span.record_exception(e)
                traceback_raw = traceback.format_exc().splitlines()
                return Execution.from_exception(e, traceback_raw)
            except Exception as e:
                span.set_status(Status(StatusCode.ERROR, str(e)))
                span.record_exception(e)
                traceback_raw = traceback.format_exc().splitlines()
                return Execution.from_exception(e, traceback_raw)

    def upload_file(self, file_path: Union[str, Path]) -> Path:
        with open(file_path, "rb") as f:
            contents = f.read()
        with open(self.remote_path / Path(file_path).name, "wb") as f:
            f.write(contents)
        _LOGGER.info(f"File ({file_path}) is uploaded to: {str(self.remote_path)}")

        return Path(self.remote_path / Path(file_path).name)

    def download_file(
        self, remote_file_path: Union[str, Path], local_file_path: Union[str, Path]
    ) -> Path:
        with open(self.remote_path / Path(remote_file_path).name, "rb") as f:
            contents = f.read()
        with open(local_file_path, "wb") as f:
            f.write(contents)
        _LOGGER.info(f"File ({remote_file_path}) is downloaded to: {local_file_path}")
        return Path(local_file_path)


class CodeInterpreterFactory:
    """Factory class for creating code interpreters.
    Could be extended to support multiple code interpreters.
    """

    _instance_map: Dict[str, CodeInterpreter] = {}
    _default_key = "default"

    @staticmethod
    def get_default_instance() -> CodeInterpreter:
        warnings.warn(
            "Use new_instance() instead for production usage, get_default_instance() is for testing and will be removed in the future."
        )
        inst_map = CodeInterpreterFactory._instance_map
        instance = inst_map.get(CodeInterpreterFactory._default_key)
        if instance:
            return instance
        instance = CodeInterpreterFactory.new_instance()
        inst_map[CodeInterpreterFactory._default_key] = instance
        return instance

    @staticmethod
    def new_instance(
        code_sandbox_runtime: Optional[str] = None,
        remote_path: Optional[Union[str, Path]] = None,
        non_exiting: bool = False,
    ) -> CodeInterpreter:
        if not code_sandbox_runtime:
            code_sandbox_runtime = os.getenv("CODE_SANDBOX_RUNTIME", "local")
        if code_sandbox_runtime == "e2b":
            envs = _get_e2b_env()
            instance: CodeInterpreter = E2BCodeInterpreter(
                timeout=_SESSION_TIMEOUT, remote_path=remote_path, envs=envs
            )
        elif code_sandbox_runtime == "local":
            instance = LocalCodeInterpreter(
                timeout=_SESSION_TIMEOUT,
                remote_path=remote_path,
                non_exiting=non_exiting,
            )
        else:
            raise ValueError(
                f"Unsupported code sandbox runtime: {code_sandbox_runtime}. Supported runtimes: e2b, local"
            )
        return instance


def _get_e2b_env() -> Union[Dict[str, str], None]:
    openai_api_key = os.getenv("OPENAI_API_KEY", "")
    anthropic_api_key = os.getenv("ANTHROPIC_API_KEY", "")
    if openai_api_key or anthropic_api_key:
        envs = {"WORKSPACE": os.getenv("WORKSPACE", "/home/user")}
        if openai_api_key:
            envs["OPENAI_API_KEY"] = openai_api_key
        if anthropic_api_key:
            envs["ANTHROPIC_API_KEY"] = anthropic_api_key
    else:
        envs = None
    return envs


def _parse_local_code_interpreter_outputs(outputs: List[Dict[str, Any]]) -> Execution:
    """Parse notebook cell outputs to Execution object. Output types:
    https://nbformat.readthedocs.io/en/latest/format_description.html#code-cell-outputs
    """
    execution = Execution()
    for data in outputs:
        if data["output_type"] == "error":
            _LOGGER.debug("Cell finished execution with error")
            execution.error = Error(
                name=data["ename"],
                value=data["evalue"],
                traceback_raw=data["traceback"],
            )
        elif data["output_type"] == "stream":
            if data["name"] == "stdout":
                execution.logs.stdout.append(data["text"])
            elif data["name"] == "stderr":
                execution.logs.stderr.append(data["text"])
        elif data["output_type"] in "display_data":
            result = Result(is_main_result=False, data=data["data"])
            execution.results.append(result)
        elif data["output_type"] == "execute_result":
            result = Result(is_main_result=True, data=data["data"])
            execution.results.append(result)
        else:
            raise ValueError(f"Unknown output type: {data['output_type']}")
    return execution


def _remove_escape_and_color_codes(input_str: str) -> str:
    pattern = re.compile(r"\x1b\[[0-9;]*[mK]")
    return pattern.sub("", input_str)



from typing import Callable, List, Optional

from .meta_tools import (
    create_code_artifact,
    edit_code_artifact,
    edit_vision_code,
    generate_vision_code,
    get_tool_descriptions,
    list_artifacts,
    object_detection_fine_tuning,
    open_code_artifact,
    use_object_detection_fine_tuning,
    view_media_artifact,
)
from .prompts import CHOOSE_PARAMS, SYSTEM_PROMPT
from .tool_utils import add_bboxes_from_masks, get_tool_descriptions_by_names
from .tools import (
    FUNCTION_TOOLS,
    TOOL_DESCRIPTIONS,
    TOOL_DOCSTRING,
    TOOLS,
    TOOLS_DF,
    TOOLS_INFO,
    UTIL_TOOLS,
    UTILITIES_DOCSTRING,
    agentic_object_detection,
    agentic_sam2_instance_segmentation,
    agentic_sam2_video_tracking,
    claude35_text_extraction,
    closest_box_distance,
    closest_mask_distance,
    countgd_object_detection,
    countgd_sam2_instance_segmentation,
    countgd_sam2_video_tracking,
    countgd_visual_prompt_object_detection,
    custom_object_detection,
    depth_anything_v2,
    detr_segmentation,
    document_extraction,
    document_qa,
    extract_frames_and_timestamps,
    florence2_object_detection,
    florence2_ocr,
    florence2_sam2_instance_segmentation,
    florence2_sam2_video_tracking,
    flux_image_inpainting,
    generate_pose_image,
    get_tool_documentation,
    load_image,
    minimum_distance,
    ocr,
    od_sam2_video_tracking,
    overlay_bounding_boxes,
    overlay_heat_map,
    overlay_segmentation_masks,
    owlv2_object_detection,
    owlv2_sam2_instance_segmentation,
    owlv2_sam2_video_tracking,
    qwen2_vl_images_vqa,
    qwen2_vl_video_vqa,
    sam2,
    save_image,
    save_json,
    save_video,
    siglip_classification,
    template_match,
    video_temporal_localization,
    vit_image_classification,
    vit_nsfw_classification,
)

__new_tools__ = [
    "import vision_agent as va",
    "from vision_agent.tools import register_tool",
]


def register_tool(imports: Optional[List] = None) -> Callable:
    def decorator(tool: Callable) -> Callable:
        import inspect

        from .tools import (  # noqa: F811
            get_tool_descriptions,
            get_tools_df,
            get_tools_info,
        )

        global TOOLS, TOOLS_DF, TOOL_DESCRIPTIONS, TOOL_DOCSTRING, TOOLS_INFO

        if tool not in TOOLS:
            TOOLS.append(tool)
            TOOLS_DF = get_tools_df(TOOLS)  # type: ignore
            TOOL_DESCRIPTIONS = get_tool_descriptions(TOOLS)  # type: ignore
            TOOL_DOCSTRING = get_tool_documentation(TOOLS)  # type: ignore
            TOOLS_INFO = get_tools_info(TOOLS)  # type: ignore

            globals()[tool.__name__] = tool
            if imports is not None:
                for import_ in imports:
                    __new_tools__.append(import_)
            __new_tools__.append(inspect.getsource(tool))
        return tool

    return decorator
